# -*- coding: utf-8 -*-
# This file is part of the 'astrophysix' Python package.
#
# Copyright © Commissariat a l'Energie Atomique et aux Energies Alternatives (CEA)
#
#  FREE SOFTWARE LICENCING
#  -----------------------
# This software is governed by the CeCILL license under French law and abiding by the rules of distribution of free
# software. You can use, modify and/or redistribute the software under the terms of the CeCILL license as circulated by
# CEA, CNRS and INRIA at the following URL: "http://www.cecill.info". As a counterpart to the access to the source code
# and rights to copy, modify and redistribute granted by the license, users are provided only with a limited warranty
# and the software's author, the holder of the economic rights, and the successive licensors have only limited
# liability. In this respect, the user's attention is drawn to the risks associated with loading, using, modifying
# and/or developing or reproducing the software by the user in light of its specific status of free software, that may
# mean that it is complicated to manipulate, and that also therefore means that it is reserved for developers and
# experienced professionals having in-depth computer knowledge. Users are therefore encouraged to load and test the
# software's suitability as regards their requirements in conditions enabling the security of their systems and/or data
# to be ensured and, more generally, to use and operate it in the same conditions as regards security. The fact that
# you are presently reading this means that you have had knowledge of the CeCILL license and that you accept its terms.
#
#
# COMMERCIAL SOFTWARE LICENCING
# -----------------------------
# You can obtain this software from CEA under other licencing terms for commercial purposes. For this you will need to
# negotiate a specific contract with a legal representative of CEA.
#
from __future__ import print_function, unicode_literals, division, absolute_import  # Python 2 and 3 compatibility

import os

from future.builtins import str
import pytest
import logging
import numpy as N

from astrophysix.simdm import Project, ProjectCategory, protocol, experiment, SimulationStudy
from astrophysix.simdm.catalogs import Catalog, TargetObject, ObjectProperty, CatalogField
from astrophysix.simdm.datafiles import PngImageFile
from astrophysix.simdm.protocol import InputParameter
from astrophysix.simdm.results import Snapshot


class TestProjectCategory(object):
    def test_from_alias_exceptions(self):
        """
        Tests exceptions raised by ProjectCategory.from_alias(method)
        """
        # Check unknown category
        with pytest.raises(ValueError) as e:
            cat = ProjectCategory.from_alias("TRICOT")
        assert str(e.value) == "No ProjectCategory defined with the alias 'TRICOT'."

    def test_init(self):
        """
        Tests valid ProjectCategory initialisation from aliases
        """
        assert ProjectCategory.from_alias("SOLAR_PHYSICS") == ProjectCategory.SolarPhysics
        assert ProjectCategory.from_alias("PLANET_ATMO") == ProjectCategory.PlanetaryAtmospheres
        assert ProjectCategory.from_alias("STAR_PLANET_INT") == ProjectCategory.StarPlanetInteractions
        assert ProjectCategory.from_alias("STAR_FORM") == ProjectCategory.StarFormation
        assert ProjectCategory.from_alias("ISM") == ProjectCategory.InterstellarMedium
        assert ProjectCategory.from_alias("MHD") == ProjectCategory.Magnetohydrodynamics
        assert ProjectCategory.from_alias("SUPERNOVAE") == ProjectCategory.Supernovae
        assert ProjectCategory.from_alias("STELLAR_ENVS") == ProjectCategory.StellarEnvironments
        assert ProjectCategory.from_alias("GALAXIES") == ProjectCategory.GalacticDynamics
        assert ProjectCategory.from_alias("COSMOLOGY") == ProjectCategory.Cosmology


class TestProject(object):
    def test_project_init(self):
        """
        Tests Project instance initialisation
        """
        # Project initialisation
        proj = Project(category=u"COSMOLOGY", project_title="My awesome project", alias="NEW_PROJ",
                       short_description="Short description of my project",
                       general_description="""This is a pretty long description for my project""",
                       data_description="The data available in this project...", acknowledgement="To cite me, ...",
                       thumbnail_image=os.path.join(os.path.dirname(__file__), "io", "datafiles", "CEA.png"),
                       directory_path="/path/to/my_data/",)
        assert isinstance(proj.project_title, str)
        assert isinstance(proj.alias, str)
        assert proj.category == ProjectCategory.Cosmology
        assert proj.project_title == "My awesome project"
        assert proj.alias == "NEW_PROJ"
        assert isinstance(proj.alias, str)
        assert proj.short_description == "Short description of my project"
        assert proj.general_description == "This is a pretty long description for my project"
        assert proj.data_description == "The data available in this project..."
        assert proj.directory_path == "/path/to/my_data/"
        assert proj.thumbnail_image == PngImageFile.load_file(os.path.join(os.path.dirname(__file__), "io",
                                                                           "datafiles", "CEA.png"))

        # Test str() conversion
        assert str(proj) == "[Cosmology] 'My awesome project' project"

    def test_setting_project_title_exc(self):
        """
        Tests that a project defined without title raises an exception
        """
        # Tests project defined without title
        with pytest.raises(AttributeError) as e:
            proj_exc = Project(category=ProjectCategory.StarFormation)  # No project title defined
        assert str(e.value) == "Project 'project_title' attribute is not defined (mandatory)."

        proj = Project(category="COSMOLOGY", project_title="My awesome project")

        # ------------- Tests invalid project title setting exception ---------------------- #
        empty_proj_title_err = "Project 'project_title' property is not a valid (non empty) string."
        with pytest.raises(AttributeError) as e:
            proj.project_title = (3, "Bananas")
        assert str(e.value) == empty_proj_title_err

        with pytest.raises(AttributeError) as e:
            proj.project_title = ""
        assert str(e.value) == empty_proj_title_err
        # ---------------------------------------------------------------------------------- #

        # Valid => should not raise any exception
        proj.project_title = "The best project of the world"

    def test_project_category_setting_exc(self):
        """
        Tests setting a project category
        """
        # Tests project defined without category
        with pytest.raises(AttributeError) as e:
            proj_exc = Project(project_title="My awesome project")  # No project category defined
        assert str(e.value) == "Project 'category' attribute is not defined (mandatory)."

        # Tests invalid category property initialisation exception
        with pytest.raises(AttributeError) as e:
            proj_exc = Project(category="TRICOT", project_title="My awesome project")  # Invalid project category
        assert str(e.value) == "No ProjectCategory defined with the alias 'TRICOT'."

        # Project initialisation
        proj = Project(category=ProjectCategory.PlanetaryAtmospheres, project_title="My awesome project")
        proj2 = Project(category="STAR_FORM", project_title="My second awesome project")

        # Tests invalid category property setting exception
        with pytest.raises(AttributeError) as e:
            proj.category = "BRODERIE"
        assert str(e.value) == "No ProjectCategory defined with the alias 'BRODERIE'."

        with pytest.raises(AttributeError) as e:
            proj.category = (3, "Strawberries")
        assert str(e.value) == "Project 'category' attribute is not a valid ProjectCategory enum value."

        # Valid => should not raise any exception
        proj.category = ProjectCategory.StarFormation

    def test_setting_project_alias_exc(self, caplog):
        """
        Tests setting a project alias

        caplog: captured log PyTest fixture
        """
        proj = Project(category="COSMOLOGY", project_title="My awesome project")

        # Tests invalid project alias setting exception
        with pytest.raises(AttributeError) as e:
            proj.alias = 25.5
        assert str(e.value) == "Project 'alias' property is not a valid string"

        # Valid => should not raise any exception
        proj.alias = "TOP_PROJ"

    def test_setting_project_long_text_fields(self):
        """
        Tests setting project short/general/data description + acknowledgement properties
        """
        proj = Project(category="COSMOLOGY", project_title="My awesome project")

        # ------------------------ Tests invalid project description setting exception ------------------------------- #
        with pytest.raises(AttributeError) as e:
            proj.short_description = 0
        assert str(e.value) == "Project 'short_description' property is not a valid string"

        # Valid => should not raise any exception
        proj.short_description = "Short description of the best project of the world"

        with pytest.raises(AttributeError) as e:
            proj.general_description = 0.1
        assert str(e.value) == "Project 'general_description' property is not a valid string"

        # Valid => should not raise any exception
        proj.general_description = "General description of the best project of the world"

        with pytest.raises(AttributeError) as e:
            proj.data_description = (23, "Monkeys")
        assert str(e.value) == "Project 'data_description' property is not a valid string"

        # Valid => should not raise any exception
        proj.data_description = "Data description of the best project of the world"

        with pytest.raises(AttributeError) as e:
            proj.acknowledgement = {"Il": "etait une fois"}
        assert str(e.value) == "Project 'acknowledgement' property is not a valid string"

        # Valid => should not raise any exception
        proj.acknowledgement = "To cite this project, please make me a coffee first..."
        # ------------------------------------------------------------------------------------------------------------ #

    def test_setting_project_directory_path(self):
        """
        Tests setting a project directory path
        """
        proj = Project(category="COSMOLOGY", project_title="My awesome project")

        with pytest.raises(AttributeError) as e:
            proj.directory_path = {'a': 45.2}
        assert str(e.value) == "Project 'directory_path' property is not a valid string"

        # Valid => should not raise any exception
        proj.directory_path = "/data/ramses/simulations/run_1/outputs/"

    def test_setting_thumbnail_image(self, tmp_path):
        """
        Tests setting a thumbnail image

        Parameters
        ----------
        tmp_path: temporary path  PyTest fixture
        """
        proj = Project(category="COSMOLOGY", project_title="My awesome project")

        # Test setting a thumbnail image with a tuple
        with pytest.raises(ValueError, match="Only image \(JPEG/PNG\) image file paths or ImageFile objects can be set "
                                             "in Project as thumbnail images."):
            proj.thumbnail_image = (4, True)

        # Try setting a . ini file as thumbnail image
        with pytest.raises(ValueError, match="Only image \(PNG/JPEG\) image file paths can be set in Project as "
                                             "thumbnail images."):
            proj.thumbnail_image = os.path.join(os.path.dirname(__file__), "io", "datafiles", "test_implode_2D.ini")

        # All good
        proj.thumbnail_image = None
        png_fpath = os.path.join(os.path.dirname(__file__), "io", "datafiles", "CEA.png")
        proj.thumbnail_image = png_fpath

        # Save study
        s = SimulationStudy(project=proj)
        study_fname = str(tmp_path / "project_thumbnail.h5")
        # Save study with a thumbnail image file
        s.save_HDF5(study_fname)

        # Reload it and save it again in same study HDF5 file with no thumbnail image defined
        ls = SimulationStudy.load_HDF5(study_fname)
        assert ls.project.thumbnail_image == PngImageFile.load_file(png_fpath)
        ls.project.thumbnail_image = None
        ls.save_HDF5(study_fname)  # new_file = False here
        ls2 = SimulationStudy.load_HDF5(study_fname)
        assert ls2.project.thumbnail_image is None

        # Relaod it and save it again in new HDF5 study file with no thumbnail image defined
        study_fname_none = str(tmp_path / "project_thumbnail_none.h5")
        ls2.save_HDF5(study_fname_none)  # new file = True here
        lsnone = SimulationStudy.load_HDF5(study_fname_none)
        assert lsnone.project.thumbnail_image is None

    def test_project_comparison(self):
        """
        Tests Project instance equality method
        """
        proj1 = Project(category=ProjectCategory.Cosmology, project_title="My awesome project", alias="NEW_PROJ",
                        short_description="Short description of my project",
                        general_description="""This is a pretty long description for my project""",
                        data_description="The data available in this project...", acknowledgement="Cite me !",
                        thumbnail_image=os.path.join(os.path.dirname(__file__), "io", "datafiles", "CEA.png"),
                        directory_path="/path/to/my_data/")

        # Different UUID => not equals
        assert proj1 != Project(category=ProjectCategory.Cosmology, project_title=proj1.project_title,
                                alias=proj1.alias, short_description=proj1.short_description,
                                general_description=proj1.general_description, data_description=proj1.data_description,
                                acknowledgement=proj1.acknowledgement, directory_path=proj1.directory_path,
                                thumbnail_image=os.path.join(os.path.dirname(__file__), "io", "datafiles", "CEA.png"))

        # Different category => not equals
        assert proj1 != Project(category=ProjectCategory.StarFormation, project_title=proj1.project_title, uid=proj1.uid,
                                alias=proj1.alias, short_description=proj1.short_description,
                                general_description=proj1.general_description, data_description=proj1.data_description,
                                acknowledgement=proj1.acknowledgement, directory_path=proj1.directory_path,
                                thumbnail_image=os.path.join(os.path.dirname(__file__), "io", "datafiles", "CEA.png"))

        # Different title => not equals
        assert proj1 != Project(category=ProjectCategory.Cosmology, project_title="My best project", uid=proj1.uid,
                                alias=proj1.alias, short_description=proj1.short_description,
                                general_description=proj1.general_description, data_description=proj1.data_description,
                                acknowledgement=proj1.acknowledgement, directory_path=proj1.directory_path,
                                thumbnail_image=os.path.join(os.path.dirname(__file__), "io", "datafiles", "CEA.png"))

        # Different alias => not equals
        assert proj1 != Project(category=ProjectCategory.Cosmology, project_title=proj1.project_title, uid=proj1.uid,
                                alias="PROJ_A", short_description=proj1.short_description,
                                general_description=proj1.general_description, data_description=proj1.data_description,
                                acknowledgement=proj1.acknowledgement, directory_path=proj1.directory_path,
                                thumbnail_image=os.path.join(os.path.dirname(__file__), "io", "datafiles", "CEA.png"))

        # Different short description => not equals
        assert proj1 != Project(category=ProjectCategory.Cosmology, project_title=proj1.project_title, uid=proj1.uid,
                                alias=proj1.alias, short_description="short project description",
                                general_description=proj1.general_description, data_description=proj1.data_description,
                                acknowledgement=proj1.acknowledgement, directory_path=proj1.directory_path,
                                thumbnail_image=os.path.join(os.path.dirname(__file__), "io", "datafiles", "CEA.png"))

        # Different general description => not equals
        assert proj1 != Project(category=ProjectCategory.Cosmology, project_title=proj1.project_title, uid=proj1.uid,
                                alias=proj1.alias, short_description=proj1.short_description,
                                general_description="gen. desc.", data_description=proj1.data_description,
                                acknowledgement=proj1.acknowledgement, directory_path=proj1.directory_path,
                                thumbnail_image=os.path.join(os.path.dirname(__file__), "io", "datafiles", "CEA.png"))

        # Different data description => not equals
        assert proj1 != Project(category=ProjectCategory.Cosmology, project_title=proj1.project_title, uid=proj1.uid,
                                alias=proj1.alias, short_description=proj1.short_description,
                                general_description=proj1.general_description, data_description="data desc",
                                acknowledgement=proj1.acknowledgement, directory_path=proj1.directory_path,
                                thumbnail_image=os.path.join(os.path.dirname(__file__), "io", "datafiles", "CEA.png"))

        # Different data directory path => not equals
        assert proj1 != Project(category=ProjectCategory.Cosmology, project_title=proj1.project_title, uid=proj1.uid,
                                alias=proj1.alias, short_description=proj1.short_description,
                                general_description=proj1.general_description, data_description=proj1.data_description,
                                acknowledgement=proj1.acknowledgement, directory_path="/other/path/",
                                thumbnail_image=os.path.join(os.path.dirname(__file__), "io", "datafiles", "CEA.png"))

        # Different project acknowledgement text => not equals
        assert proj1 != Project(category=ProjectCategory.Cosmology, project_title=proj1.project_title, uid=proj1.uid,
                                alias=proj1.alias, short_description=proj1.short_description,
                                general_description=proj1.general_description, data_description=proj1.data_description,
                                acknowledgement="Hey ! This is a good citation...", directory_path=proj1.directory_path,
                                thumbnail_image=os.path.join(os.path.dirname(__file__), "io", "datafiles", "CEA.png"))

        # Different project thumbnail images
        assert proj1 != Project(category=ProjectCategory.Cosmology, project_title=proj1.project_title, uid=proj1.uid,
                                alias=proj1.alias, short_description=proj1.short_description,
                                general_description=proj1.general_description, data_description=proj1.data_description,
                                acknowledgement=proj1.acknowledgement, directory_path=proj1.directory_path,
                                thumbnail_image=os.path.join(os.path.dirname(__file__), "io",
                                                             "datafiles", "irfu_simple.jpg"))

        # Identical projects
        p = Project(category=ProjectCategory.Cosmology, project_title=proj1.project_title, uid=proj1.uid,
                    alias=proj1.alias, short_description=proj1.short_description, directory_path=proj1.directory_path,
                    general_description=proj1.general_description, data_description=proj1.data_description,
                    thumbnail_image=os.path.join(os.path.dirname(__file__), "io", "datafiles", "CEA.png"),
                    acknowledgement=proj1.acknowledgement)
        assert proj1 == p

        # Simulation list differ => not equals
        ramses = protocol.SimulationCode(name="Ramses 3 (MHD)", code_name="Ramses")
        proj1.simulations.add(experiment.Simulation(name="My simu épatante", alias="SIMU_1", simu_code=ramses))
        assert proj1 != p

    def test_project_invalid_attribute(self, caplog):
        """
        Tests project invalid attribute warning log message

        Parameters
        ----------
        caplog: captured log PyTest fixture
        """
        proj = Project(category="COSMOLOGY", project_title="Proj title",
                       attribute_A=34.619)
        assert caplog.record_tuples[-1] == ('astrophysix', logging.WARNING,
                                            "'attribute_A' attribute is unknown for a 'Project' object. It is therefore"
                                            " ignored.")

    def test_project_simu_datatable_insertion_deletion(self):
        """

        :return:
        """
        """
        Tests project Galactica validity checks

        Parameters
        ----------
        caplog: captured log PyTest fixture
        """
        proj = Project(category="COSMOLOGY", project_title="Proj title", alias="TEST_PROJ")
        ramses = protocol.SimulationCode(name="Ramses 3.0 (Hydro)", code_name="RAMSES", alias="RAMSES_CODE")
        # Add input parameters
        lmin = ramses.input_parameters.add(InputParameter(key="levelmin", name="l_min", description="min. level of AMR refinement"))
        lmax = ramses.input_parameters.add(InputParameter(key="levelmax", name="l_max", description="max. level of AMR refinement"))
        boxlen = InputParameter(key="boxlen", name="L_box", description="size of simulation domain")

        proj.simulations.add(experiment.Simulation(ramses, name="My simu (hydro)", alias="SIMU_1",
                                                   execution_time="2020-11-01 10:30:00"))

        with pytest.raises(AttributeError, match="Project '\[boxlen\] 'L_box' input parameter' does not refer to any "
                                                 "input parameter of the project's simulation codes."):
            proj.simu_datatable_params.add(boxlen)  # Boxlen is not an input parameter of RAMSES

        proj.simu_datatable_params.add(lmin)
        proj.simu_datatable_params.add(lmax)
        with pytest.raises(AttributeError, match="'\[levelmin\] 'l_min' input parameter' cannot be deleted, the "
                                                 "following items depend on it \(try to delete them first\) : "
                                                 "\[Simulation datatable \[levelmin\] 'l_min' input parameter\]."):
            del ramses.input_parameters[lmin]  # lmin belongs to the list of simulation datatable parameters, you cannot delete it

        # Delete it first from the simu datatable parameter list, then delete it from RAMSES input parameter list
        del proj.simu_datatable_params[lmin]
        del ramses.input_parameters[lmin]  # Ok

    def test_project_galactica_validity_checks(self, caplog):
        """
        Tests project Galactica validity checks

        Parameters
        ----------
        caplog: captured log PyTest fixture
        """
        proj = Project(category="COSMOLOGY", project_title="Proj title",
                       short_description="This is short description of my project")

        # No alias defined
        proj.galactica_validity_check()
        assert caplog.record_tuples[-1] == ('astrophysix.simdm', logging.WARNING,
                                            "[Cosmology] 'Proj title' project Galactica alias is missing.")

        # Alias too long
        proj.alias = "MY_ALIAS_123456789_10"
        proj.galactica_validity_check()
        assert caplog.record_tuples[-1] == ('astrophysix.simdm', logging.WARNING,
                                            "[Cosmology] 'Proj title' project Galactica alias is too long (max. 16 "
                                            "characters).")
        proj.alias = "MY_ALIAS"  # => Ok

        # Invalid Galactica aliases
        for inv_alias in ["8PROJ", "PROJ+SCIENCE", "_PROJ_A", "PROJ_", "PROJ_hydro", "hydro_RUN_4", "MY_hydro_8", "B_"]:
            proj.alias = inv_alias
            proj.galactica_validity_check()
            assert caplog.record_tuples[-1] == ('astrophysix.simdm', logging.WARNING,
                                                "[Cosmology] 'Proj title' project Galactica alias is not valid (The "
                                                "alias can contain capital letters, digits and '_' only. It must start "
                                                "with a capital letter and cannot end with a '_'.)")
            proj.alias = "MY_ALIAS_123456789_10"  # Alias too long => just to interleave another warning message
            proj.galactica_validity_check()
        proj.alias = "MY_ALIAS"  # => Ok

        # No project title defined => Should never happen
        # [...]

        # Project title too long
        proj.project_title = "This  is a pretty long title for a project that has absolutely no sense. It means that " \
                             "this project title can be as long as 400 hundred chars and still be valid ? Come on..."
        proj.galactica_validity_check()
        assert caplog.record_tuples[-1] == ('astrophysix.simdm', logging.WARNING,
                                            "[Cosmology] 'This  is a pretty long title for a project that has "
                                            "absolutely no sense. It means that this project title can be as long as "
                                            "400 hundred chars and still be valid ? Come on...' project Galactica "
                                            "project title is too long (max. 128 characters).")
        proj.project_title = "Proj title"  # => Ok

        # No short description defined
        proj.short_description = ""
        proj.galactica_validity_check()
        assert caplog.record_tuples[-1] == ('astrophysix.simdm', logging.WARNING,
                                            "[Cosmology] 'Proj title' project Galactica short description is missing.")

        # Short description too long
        proj.short_description = "zzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzz" \
                                 "yyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyy" \
                                 "xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx" \
                                 "wwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwwww"
        proj.galactica_validity_check()
        assert caplog.record_tuples[-1] == ('astrophysix.simdm', logging.WARNING,
                                            "[Cosmology] 'Proj title' project Galactica short description is too long "
                                            "(max. 256 characters).")
        proj.short_description = "This is a pretty short description"  # => Ok

        # Simulation code duplicates (same aliases)
        ramses1 = protocol.SimulationCode(name="Ramses 3.0 (Hydro)", code_name="RAMSES", alias="RAMSES_CODE")
        ramses2 = protocol.SimulationCode(name="Ramses 3.1 (MHD)", code_name="RAMSES", alias="RAMSES_CODE")
        simu1 = proj.simulations.add(experiment.Simulation(ramses1, name="My simu (hydro)", alias="SIMU_1",
                                                           execution_time="2020-11-01 10:30:00"))
        simu2 = proj.simulations.add(experiment.Simulation(ramses2, name="My simu (MHD)", alias="SIMU_2",
                                                           execution_time="2020-11-01 10:30:00"))
        proj.galactica_validity_check()
        assert caplog.record_tuples[-1] == ('astrophysix.simdm', logging.WARNING,
                                            "[Ramses 3.0 (Hydro)] simulation code and [Ramses 3.1 (MHD)] simulation "
                                            "code protocols share the same alias. They must be unique.")
        ramses2.alias = "RAMSES_OTHER"  # protocol.alias differ => Ok

        # Simulation duplicates (same aliases)
        simu2.alias = simu1.alias
        proj.galactica_validity_check()
        assert caplog.record_tuples[-1] == ('astrophysix.simdm', logging.WARNING,
                                            "'My simu (hydro)' simulation and 'My simu (MHD)' simulation experiments "
                                            "share the same alias. They must be unique.")
        simu1.alias = "MY_ALIAS_SIMU"

        # Tests target object duplicates (same name)
        sn1 = simu1.snapshots.add(Snapshot(name="My snapshot"))
        sn2 = simu2.snapshots.add(Snapshot(name="My snapshot (2)"))
        tobj = TargetObject(name="Galaxy")
        m = tobj.object_properties.add(ObjectProperty(property_name="mass"))
        cat1 = sn1.catalogs.add(Catalog(name="My catalog", target_object=tobj))
        cat1.catalog_fields.add(CatalogField(obj_prop=m, values=N.random.uniform(size=10)))
        tobj2 = TargetObject(name="Galaxy")
        m2 = tobj2.object_properties.add(ObjectProperty(property_name="mass"))
        cat2 = sn2.catalogs.add(Catalog(name="My catalog", target_object=tobj2))
        cat2.catalog_fields.add(CatalogField(obj_prop=m2, values=N.random.uniform(size=10)))
        proj.galactica_validity_check()
        assert caplog.record_tuples[-1] == ('astrophysix.simdm', logging.WARNING,
                                            "'Galaxy' target object and 'Galaxy' target object share the same name. "
                                            "They must be unique.")

    def test_project_targobj_hdf5_io_compatibility_v1(self):
        """Backward compatibility test : tests loading a Simulation study saved with version v<0.5.0 where Catalog and
        TargetObject instances did not exist."""
        # Try to load a Project with no catalog/target object (v1), saved by an older version of astrophysix (v<0.5)
        study_0_4_2_path = os.path.join(os.path.dirname(__file__), "io", "backward_compat", "study_v0.4.2.h5")
        study = SimulationStudy.load_HDF5(study_0_4_2_path)
        assert study.project.project_title == "My awesome project"
        assert study.project.acknowledgement == ""  # No 'acknowledgement' property defined in v0.4.2
        assert study.project.thumbnail_image is None  # No 'thumbnail_image' property defined in v0.6
        # No target object
        assert len([tobj for tobj in study.project._target_objects()]) == 0

    def test_project_hdf5_io(self, tmp_path):
        """
        Tests saving/loading Project from HDF5 file

        Parameters
        ----------
        tmp_path: temporary path  PyTest fixture
        """
        proj = Project(category="COSMOLOGY", project_title="My awesome project€", alias="NEW_PROJ",
                       short_description="Short description of my project",
                       general_description="""This is a pretty long description for my project""",
                       data_description="The data available in this project...", acknowledgement="Cite me !",
                       thumbnail_image=os.path.join(os.path.dirname(__file__), "io", "datafiles", "irfu_simple.jpg"),
                       directory_path="/path/to/my_data/")
        ramses = protocol.SimulationCode(name="Ramses 3 (MHD)", code_name="Ramses", code_version="3.0.0")
        proj.simulations.add(experiment.Simulation(name="My simu épatante", alias="SIMU_1",
                                                   description="simu description", simu_code=ramses))

        fname = str(tmp_path / "proj.h5")
        study = SimulationStudy(project=proj)
        study.save_HDF5(fname)
        study_loaded = SimulationStudy.load_HDF5(fname)
        proj_loaded = study_loaded.project
        assert proj_loaded == proj


__all__ = ["TestProjectCategory", "TestProject"]




# -*- coding: utf-8 -*-
# This file is part of the 'astrophysix' Python package.
#
# Copyright © Commissariat a l'Energie Atomique et aux Energies Alternatives (CEA)
#
#  FREE SOFTWARE LICENCING
#  -----------------------
# This software is governed by the CeCILL license under French law and abiding by the rules of distribution of free
# software. You can use, modify and/or redistribute the software under the terms of the CeCILL license as circulated by
# CEA, CNRS and INRIA at the following URL: "http://www.cecill.info". As a counterpart to the access to the source code
# and rights to copy, modify and redistribute granted by the license, users are provided only with a limited warranty
# and the software's author, the holder of the economic rights, and the successive licensors have only limited
# liability. In this respect, the user's attention is drawn to the risks associated with loading, using, modifying
# and/or developing or reproducing the software by the user in light of its specific status of free software, that may
# mean that it is complicated to manipulate, and that also therefore means that it is reserved for developers and
# experienced professionals having in-depth computer knowledge. Users are therefore encouraged to load and test the
# software's suitability as regards their requirements in conditions enabling the security of their systems and/or data
# to be ensured and, more generally, to use and operate it in the same conditions as regards security. The fact that
# you are presently reading this means that you have had knowledge of the CeCILL license and that you accept its terms.
#
#
# COMMERCIAL SOFTWARE LICENCING
# -----------------------------
# You can obtain this software from CEA under other licencing terms for commercial purposes. For this you will need to
# negotiate a specific contract with a legal representative of CEA.
#
from __future__ import print_function, unicode_literals, division, absolute_import  # Python 2 and 3 compatibility
from future.builtins import str, dict
import pytest
import logging

from astrophysix.simdm import ProjectCategory, Project, SimulationStudy
from astrophysix.simdm.experiment import Simulation
from astrophysix.simdm.protocol import AlgoType, Algorithm, Physics, PhysicalProcess, InputParameter, SimulationCode,\
    PostProcessingCode
from astrophysix.simdm.protocol.base import Protocol


class TestAlgoType(object):
    def test_from_key_exceptions(self):
        """
        Tests exceptions raised by AlgoType.from_alias(method)
        """
        # Check unknown algorithm type
        with pytest.raises(ValueError) as e:
            at = AlgoType.from_key("MENUISERIE")
        assert str(e.value) == "No AlgoType defined with the key 'MENUISERIE'."

    def test_init(self):
        """
        Tests valid AlgoType initialisation from aliases
        """
        assert AlgoType.from_key("PM") == AlgoType.ParticleMesh
        assert AlgoType.from_key("Poisson_MG") == AlgoType.PoissonMultigrid
        assert AlgoType.from_key("Godunov") == AlgoType.Godunov
        assert AlgoType.from_key("HLLC") == AlgoType.HLLCRiemann
        assert AlgoType.from_key("AMR") == AlgoType.AdaptiveMeshRefinement
        assert AlgoType.from_key("Voronoi_MM") == AlgoType.VoronoiMovingMesh
        assert AlgoType.from_key("FOF") == AlgoType.FriendOfFriend
        assert AlgoType.from_key("Poisson_CG") == AlgoType.PoissonConjugateGradient
        assert AlgoType.from_key("SPH") == AlgoType.SmoothParticleHydrodynamics
        assert AlgoType.from_key("ray_tracer") == AlgoType.RayTracer


class TestPhysics(object):
    def test_from_key_exceptions(self):
        """
        Tests exceptions raised by Physics.from_alias(method)
        """
        # Check unknown physical process
        with pytest.raises(ValueError) as e:
            ph = Physics.from_key("Plomberie")
        assert str(e.value) == "No Physics defined with the key 'Plomberie'."

    def test_init(self):
        """
        Tests valid Physics initialisation from aliases
        """
        assert Physics.from_key("sn_feedback") == Physics.SupernovaeFeedback
        assert Physics.from_key("self_gravity") == Physics.SelfGravity
        assert Physics.from_key("hydro") == Physics.Hydrodynamics
        assert Physics.from_key("mhd") == Physics.MHD
        assert Physics.from_key("mol_cooling") == Physics.MolecularCooling
        assert Physics.from_key("star_form") == Physics.StarFormation
        assert Physics.from_key("AGN_feedback") == Physics.AGNFeedback


class TestAlgorithm(object):
    def test_algo_init(self):
        """
        Tests Algorithm instance initialisation
        """
        # Algorithm initialisation
        amr = Algorithm(algo_type=AlgoType.AdaptiveMeshRefinement, description="Adaptive mesh refinement implementation")

        assert amr.name == AlgoType.AdaptiveMeshRefinement.name
        assert amr.description == "Adaptive mesh refinement implementation"

        # Test str() conversion
        assert str(amr) == "'Adaptive mesh refinement' algorithm"

    def test_setting_algorithm_type_exc(self):
        """
        Tests that an algorithm defined without type raises an exception
        """
        # Tests algorithm defined without type
        with pytest.raises(AttributeError) as e:
            algo_exc = Algorithm()  # No algorithm type defined
        assert str(e.value) == "Algorithm 'algo_type' attribute is not defined (mandatory)."

        # Tests invalid algo type property initialisation exception
        with pytest.raises(AttributeError) as e:
            algo_exc = Algorithm(algo_type="CUISINE")  # Invalid algorithm type
        assert str(e.value) == "No AlgoType defined with the key 'CUISINE'."

        with pytest.raises(AttributeError) as e:
            algo_exc = Algorithm(algo_type=(345, "Elephant"))  # Invalid algorithm type
        assert str(e.value) == "Algorithm 'algo_type' attribute is not a valid AlgoType enum value."

    def test_setting_algorithm_description(self):
        """
        Tests setting algorithm description property
        """
        algo = Algorithm(algo_type=AlgoType.Godunov, description="Descr of algo")

        # ------------------------ Tests invalid algorithm description setting exception ----------------------------- #
        with pytest.raises(AttributeError) as e:
            algo.description = 0
        assert str(e.value) == "Algorithm 'description' property is not a valid string."

        # Valid => should not raise any exception
        algo.description = "Description of the best algorithm in the world"
        # ------------------------------------------------------------------------------------------------------------ #

    def test_algorithm_comparison(self):
        """
        Tests rich comparison method Algorithm.__eq__()
        """
        algo = Algorithm(algo_type=AlgoType.ParticleMesh, description="My algo descr.")

        # Different UUID => not equals
        assert algo != Algorithm(algo_type=algo.algo_type, description=algo.description)

        # Different type => not equals
        assert algo != Algorithm(algo_type=AlgoType.AdaptiveMeshRefinement, uid=algo.uid, description=algo.description)

        # Different description => not equals
        assert algo != Algorithm(algo_type=algo.algo_type, uid=algo.uid, description="Other descr.")

        # Identical algorithms
        assert algo == Algorithm(algo_type=algo.algo_type, uid=algo.uid, description=algo.description)


class TestInputParameter(object):
    def test_inpparam_init(self):
        """
        Tests InputParameter instance initialisation
        """
        # Input parameter initialisation
        lmin = InputParameter(key="levelmin", name="lmin", description="Min. level of AMR refinement")

        assert lmin.key == "levelmin"
        assert lmin.name == "lmin"
        assert lmin.description == "Min. level of AMR refinement"

        # Test str() conversion
        assert str(lmin) == "[levelmin] 'lmin' input parameter"

    def test_setting_inpparam_name_exc(self):
        """
        Tests that an input parameter defined without a name raises an exception
        """
        # Tests input parameter defined without name
        with pytest.raises(AttributeError) as e:
            ip_exc = InputParameter(key="ngridmax")  # No input parameter name defined
        assert str(e.value) == "Input parameter 'name' attribute is not defined (mandatory)."

        # Tests input parameter name property initialisation exception
        invalid_inpparam_name_err = "Input parameter 'name' property is not a valid \\(non empty\\) string."
        with pytest.raises(AttributeError, match=invalid_inpparam_name_err):
            ip_exc = InputParameter(name=-4)  # Invalid input parameter name

        with pytest.raises(AttributeError, match=invalid_inpparam_name_err):
            ip_exc = InputParameter(name="")  # Invalid input parameter name (empty string)

        ip = InputParameter(name="a")
        with pytest.raises(AttributeError, match=invalid_inpparam_name_err):
            ip.name = -1  # Invalid input parameter name

    def test_setting_inpparam_description(self):
        """
        Tests setting input parameter description property
        """
        ip = InputParameter(name="x", description="Descr of input parameter")

        # ------------------------ Tests invalid input parameter description setting exception ----------------------- #
        invalid_inpparam_descr_err = "Input parameter 'description' property is not a valid string."
        with pytest.raises(AttributeError) as e:
            ip.description = 0
        assert str(e.value) == invalid_inpparam_descr_err

        with pytest.raises(AttributeError) as e:
            ip_exc = InputParameter(name="x", description=("Blue", "turtle"))
        assert str(e.value) == invalid_inpparam_descr_err
        # ------------------------------------------------------------------------------------------------------------ #

        # Valid => should not raise any exception
        ip.description = "Description of the best algorithm in the world"

    def test_inputparam_comparison(self):
        """
        Tests rich comparison method InputParameter.__eq__()
        """
        lmin = InputParameter(key="levelmin", name="lmin", description="Min. level of AMR refinement")

        # Different UUID => not equals
        assert lmin != InputParameter(key=lmin.key, name=lmin.name, description=lmin.description)

        # Different key => not equals
        assert lmin != InputParameter(key="other_key", uid=lmin.uid, name=lmin.name, description=lmin.description)

        # Different name => not equals
        assert lmin != InputParameter(key=lmin.key, uid=lmin.uid, name="other_name", description=lmin.description)

        # Different description => not equals
        assert lmin != InputParameter(key=lmin.key, uid=lmin.uid, name=lmin.name, description="Alt. description")

        # Identical input parameter
        assert lmin == InputParameter(key=lmin.key, uid=lmin.uid, name=lmin.name, description=lmin.description)

    def test_inpparam_galactica_validity_checks(self, caplog):
        """
        Tests input parameter Galactica validity checks

        Parameters
        ----------
        caplog: captured log PyTest fixture
        """
        lmin = InputParameter(key="levelmin", name="lmin")

        # No input parameter name defined => Should never happen
        # [...]

        # Input parameter name too long
        lmin.name = "This is a much too long name for a basic input parameter of the RAMSES code"
        lmin.galactica_validity_check()
        assert caplog.record_tuples[-1] == ('astrophysix.simdm', logging.WARNING,
                                            "[levelmin] 'This is a much too long name for a basic input parameter of "
                                            "the RAMSES code' input parameter Galactica input parameter name is too "
                                            "long (max. 64 characters).")
        lmin.name = "lmin"  # => Ok

        # input parameter key too long
        lmin.key = "my_too_long_input_param_key"
        lmin.galactica_validity_check()
        assert caplog.record_tuples[-1] == ('astrophysix.simdm', logging.WARNING,
                                            "[my_too_long_input_param_key] 'lmin' input parameter Galactica input "
                                            "parameter key is too long (max. 16 characters).")
        lmin.key = "levelmin"  # => Ok


class TestPhysicalProcess(object):
    def test_physics_init(self):
        """
        Tests PhysicalProcess instance initialisation
        """
        # Physical process initialisation
        agn_fb = PhysicalProcess(physics=Physics.AGNFeedback, description="feedback from Active Galactic Nuclei")

        assert agn_fb.name == Physics.AGNFeedback.name
        assert agn_fb.description == "feedback from Active Galactic Nuclei"

        # Test str() conversion
        assert str(agn_fb) == "'AGN feedback' physical process"

    def test_setting_physproc_physics_exc(self):
        """
        Tests that a physical process defined without physics raises an exception
        """
        # Tests physical process defined without physics
        with pytest.raises(AttributeError) as e:
            ph_exc = PhysicalProcess(description="simple descr")  # No physics defined
        assert str(e.value) == "PhysicalProcess 'physics' attribute is not defined (mandatory)."

        # Tests invalid physical process 'physics' property initialisation exception
        with pytest.raises(AttributeError) as e:
            ph_exc = PhysicalProcess(physics="Wind")  # Invalid physics
        assert str(e.value) == "No Physics defined with the key 'Wind'."

        with pytest.raises(AttributeError) as e:
            ph_exc = PhysicalProcess(physics=(47, "Zebras"))  # Invalid physics
        assert str(e.value) == "PhysicalProcess 'physics' attribute is not a valid Physics enum value."

    def test_setting_physproc_description(self):
        """
        Tests setting physical process description property
        """
        ph = PhysicalProcess(physics=Physics.Hydrodynamics, description="Descr of process")

        # --------------------- Tests invalid physical process description setting exception ------------------------- #
        with pytest.raises(AttributeError) as e:
            ph.description = -785.2
        assert str(e.value) == "PhysicalProcess 'description' property is not a valid string."

        # Valid => should not raise any exception
        ph.description = "Description of the collisional physics"
        # ------------------------------------------------------------------------------------------------------------ #

    def test_physical_process_comparison(self):
        """
        Tests rich comparison method PhysicalProcess.__eq__()
        """
        agn_fb = PhysicalProcess(physics=Physics.AGNFeedback, description="feedback from Active Galactic Nuclei")

        # Different UUID => not equals
        assert agn_fb != PhysicalProcess(physics=agn_fb.physics, description=agn_fb.description)

        # Different type => not equals
        assert agn_fb != PhysicalProcess(physics=Physics.StarFormation, uid=agn_fb.uid, description=agn_fb.description)

        # Identical physical process
        assert agn_fb == PhysicalProcess(physics=agn_fb.physics, uid=agn_fb.uid, description=agn_fb.description)


class TestProtocol(object):
    def test_protocol_init(self):
        """
        Tests Protocol instance initialisation
        """
        # Protocol initialisation
        prot = Protocol(name="Ramses 3", code_name="RAMSES", code_version="3.10.0", alias="RAMSES_3",
                        url="http://www.my-code.org", description="Code description")
        assert prot.name == "Ramses 3"
        assert prot.code_name == "RAMSES"
        assert prot.code_version == "3.10.0"
        assert prot.alias == "RAMSES_3"
        assert prot.description == "Code description"
        assert prot.url == "http://www.my-code.org"

        # Test str() conversion
        assert str(prot) == "[Ramses 3]"

    def test_setting_protocol_name_exc(self):
        """
        Tests that a protocol defined without name raises an exception
        """
        # Tests protocol defined without name
        with pytest.raises(AttributeError) as e:
            prot_exc = Protocol(code_name="RAMSES")  # No Protocol code name defined
        assert str(e.value) == "Protocol 'name' attribute is not defined (mandatory)."

        code = Protocol(name="Ramses 3", code_name="RAMSES")

        # ------------- Tests invalid protocol name  setting exception ---------------------- #
        empty_protocol_name_err = "Protocol 'name' property is not a valid (non empty) string."
        with pytest.raises(AttributeError) as e:
            code.name = ("Cat", 20)
        assert str(e.value) == empty_protocol_name_err

        with pytest.raises(AttributeError) as e:
            code.name = ""
        assert str(e.value) == empty_protocol_name_err
        # ----------------------------------------------------------------------------------- #

        # Valid => should not raise any exception
        code.name = "The best exascale code on Earth"

    def test_setting_protocol_codename_exc(self):
        """
        Tests that a protocol defined without code name raises an exception
        """
        # Tests protocol defined without code name
        with pytest.raises(AttributeError) as e:
            proto_exc = Protocol(name="Ramses 3")  # No protocol code name defined
        assert str(e.value) == "Protocol 'code_name' attribute is not defined (mandatory)."

        code = Protocol(name="Ramses 3", code_name="RAMSES")

        # ------------- Tests invalid protocol name  setting exception ---------------------- #
        empty_protocol_codename_err = "Protocol 'code_name' property is not a valid (non empty) string."
        with pytest.raises(AttributeError) as e:
            code.code_name = (2, "Dolphin")
        assert str(e.value) == empty_protocol_codename_err

        with pytest.raises(AttributeError) as e:
            code.code_name = ""
        assert str(e.value) == empty_protocol_codename_err
        # ----------------------------------------------------------------------------------- #

        # Valid => should not raise any exception
        code.code_name = "GADGET"

    def test_setting_protocol_codeversion_exc(self):
        """
        Tests setting a protocol code version
        """
        code = Protocol(name="Ramses 3", code_name="RAMSES")

        # ------------- Tests invalid protocol code version setting exception ---------------------- #
        invalid_protocol_codeversion_err = "Protocol 'code_version' property is not a valid string."
        with pytest.raises(AttributeError) as e:
            code.code_version = -4.5
        assert str(e.value) == invalid_protocol_codeversion_err
        # ------------------------------------------------------------------------------------------ #

        # Valid => should not raise any exception
        code.code_version = "GADGET"

    def test_setting_protocol_url_exc(self):
        """
        Tests setting a protocol url
        """
        code = Protocol(name="Ramses 3", code_name="RAMSES")

        # ------------- Tests invalid protocol url setting exception ---------------------- #
        invalid_protocol_url_err = "Protocol 'url' property is not a valid string."
        with pytest.raises(AttributeError) as e:
            code.url = 45
        assert str(e.value) == invalid_protocol_url_err
        # --------------------------------------------------------------------------------- #

        # Valid => should not raise any exception
        code.url = "1.2.3"

    def test_setting_protocol_alias_exc(self, caplog):
        """
        Tests setting a protocol alias

        Parameters
        ----------
        caplog: captured log PyTest fixture
        """
        code = Protocol(name="Ramses 3", code_name="RAMSES", alias="RAMSES_3")

        # Tests invalid protocol alias setting exception
        with pytest.raises(AttributeError) as e:
            code.alias = -1.245
        assert str(e.value) == "Protocol 'alias' property is not a valid string"

        # Valid => should not raise any exception
        code.alias = "TOP_CODE"

    def test_setting_protocol_description(self):
        """
        Tests setting protocol description property
        """
        code = Protocol(name="Ramses 3", code_name="RAMSES")

        # ------------------------ Tests invalid protocol description setting exception ------------------------------ #
        with pytest.raises(AttributeError) as e:
            code.description = 0
        assert str(e.value) == "Protocol 'description' property is not a valid string."

        # Valid => should not raise any exception
        code.description = "Description of the best code in the world"
        # ------------------------------------------------------------------------------------------------------------ #

    def test_protocol_equality(self):
        """
        Tests rich comparison method Protocol.__eq__()
        """
        protocol = Protocol(name="Ramses 3 (MHD)", code_name="Ramses", code_version="3.10.1", alias="RAMSES_3",
                            url="http://www.my-code.net", description="This is a fair description")

        # Different UUID => not equals
        assert protocol != Protocol(name=protocol.name, code_name=protocol.code_name,
                                    code_version=protocol.code_version, alias=protocol.alias, url=protocol.url,
                                    description=protocol.description)

        # Different name => not equals
        assert protocol != Protocol(name="Ramses 3",  uid=protocol.uid, code_name=protocol.code_name,
                                    code_version=protocol.code_version, alias=protocol.alias, url=protocol.url,
                                    description=protocol.description)

        # Different code name => not equals
        assert protocol != Protocol(name=protocol.name,  uid=protocol.uid, code_name="Gadget",
                                    code_version=protocol.code_version, alias=protocol.alias, url=protocol.url,
                                    description=protocol.description)

        # Different code version => not equals
        assert protocol != Protocol(name=protocol.name,  uid=protocol.uid, code_name=protocol.code_name,
                                    code_version="3.10.2", alias=protocol.alias, url=protocol.url,
                                    description=protocol.description)

        # Different alias => not equals
        assert protocol != Protocol(name=protocol.name,  uid=protocol.uid, code_name=protocol.code_name,
                                    code_version=protocol.code_version, alias="RAMSES_3.10.2", url=protocol.url,
                                    description=protocol.description)

        # Different url => not equals
        assert protocol != Protocol(name=protocol.name,  uid=protocol.uid, code_name=protocol.code_name,
                                    code_version=protocol.code_version, alias=protocol.alias, url="http://irfu.cea.fr",
                                    description=protocol.description)

        # Different descritpion => not equals
        assert protocol != Protocol(name=protocol.name,  uid=protocol.uid, code_name=protocol.code_name,
                                    code_version=protocol.code_version, alias=protocol.alias, url=protocol.url,
                                    description="Best description")

        # Identical protocols
        p = Protocol(name=protocol.name, uid=protocol.uid, code_name=protocol.code_name,
                     code_version=protocol.code_version, alias=protocol.alias, url=protocol.url,
                     description=protocol.description)
        assert protocol == p

        # Add algorithms
        a1 = protocol.algorithms.add(Algorithm(algo_type=AlgoType.AdaptiveMeshRefinement, description="AMR descr"))
        a2 = protocol.algorithms.add(Algorithm(algo_type=AlgoType.Godunov, description="Godunov scheme"))
        a3 = protocol.algorithms.add(Algorithm(algo_type=AlgoType.HLLCRiemann, description="HLLC Riemann solver"))
        a4 = protocol.algorithms.add(Algorithm(algo_type=AlgoType.PoissonMultigrid,
                                               description="Multigrid Poisson solver"))
        a5 = protocol.algorithms.add(Algorithm(algo_type=AlgoType.ParticleMesh, description="PM solver"))

        # Protocol algo list differ => not equals
        p.algorithms.add(a1)
        p.algorithms.add(a2)
        p.algorithms.add(a3)
        p.algorithms.add(a4)
        assert p != protocol
        # Same algorithms => identical protocols
        p.algorithms.add(a5)
        assert p == protocol

        # Add input parameters
        ip1 = protocol.input_parameters.add(InputParameter(key="levelmin", name="levelmin",
                                                           description="min. level of AMR refinement"))
        ip2 = protocol.input_parameters.add(InputParameter(key="levelmax", name="levelmax",
                                                           description="max. level of AMR refinement"))

        # Protocol input parameter list differ => not equals
        p.input_parameters.add(ip1)
        assert p != protocol

        # Same input parameters => identical protocols
        p.input_parameters.add(ip2)
        assert p == protocol

    def test_protocol_galactica_validity_checks(self, caplog):
        """
        Tests protocol Galactica validity checks

        Parameters
        ----------
        caplog: captured log PyTest fixture
        """
        protocol = Protocol(name="Ramses 3 (MHD)", code_name="RAMSES", code_version="3.10.1")

        # No alias defined
        protocol.galactica_validity_check()
        assert caplog.record_tuples[-1] == ('astrophysix.simdm', logging.WARNING,
                                            "[Ramses 3 (MHD)] Galactica protocol alias is missing.")

        # Alias too long
        protocol.alias = "MY_ALIAS_123456789_10"
        protocol.galactica_validity_check()
        assert caplog.record_tuples[-1] == ('astrophysix.simdm', logging.WARNING,
                                            "[Ramses 3 (MHD)] Galactica protocol alias is too long (max. 16 "
                                            "characters).")
        protocol.alias = "MY_ALIAS"  # => Ok

        # Invalid Galactica alias
        protocol.alias = "invalid_ALIAS"  # non capital letters !
        protocol.galactica_validity_check()
        assert caplog.record_tuples[-1] == ('astrophysix.simdm', logging.WARNING,
                                            "[Ramses 3 (MHD)] Galactica protocol alias is not valid (The alias can "
                                            "contain capital letters, digits and '_' only. It must start with a capital"
                                            " letter and cannot end with a '_'.)")
        protocol.alias = "MY_ALIAS"  # => Ok

        # No protocol name defined => Should never happen
        # [...]

        # Protocol name too long
        protocol.name = "This  is a pretty long name for a prototol that has absolutely no sense. It means that " \
                        "this protocol can be as long as 400 hundred chars and still be valid ? Come on..."
        protocol.galactica_validity_check()
        assert caplog.record_tuples[-1] == ('astrophysix.simdm', logging.WARNING,
                                            "[This  is a pretty long name for a prototol that has absolutely no sense. "
                                            "It means that this protocol can be as long as 400 hundred chars and still "
                                            "be valid ? Come on...] Galactica protocol name is too long (max. 128 "
                                            "characters).")
        protocol.name = "Ramses 3 (MHD)"  # => Ok

        # No protocol code name defined => Should never happen
        # [...]

        # Protocol code name too long
        protocol.code_name = "CODE Ultimate with a much too long name"
        protocol.galactica_validity_check()
        assert caplog.record_tuples[-1] == ('astrophysix.simdm', logging.WARNING,
                                            "[Ramses 3 (MHD)] Galactica protocol code name is too long (max. 32 "
                                            "characters).")
        protocol.code_name = "RAMSES"  # => Ok

        # No code version defined
        protocol.code_version = ""
        protocol.galactica_validity_check()
        assert caplog.record_tuples[-1] == ('astrophysix.simdm', logging.WARNING,
                                            "[Ramses 3 (MHD)] Galactica protocol code version is missing.")

        # Code version too long
        protocol.code_version = "145.1254.3rc20"
        protocol.galactica_validity_check()
        assert caplog.record_tuples[-1] == ('astrophysix.simdm', logging.WARNING,
                                            "[Ramses 3 (MHD)] Galactica protocol code version is too long (max. 8 "
                                            "characters).")
        protocol.code_version = "3.10.1"  # => Ok


class TestPostProcessingCode(object):
    def test_postprocessor_init(self):
        """
        Tests PostProcessingCode instance initialisation
        """
        ramses = PostProcessingCode(name="Ramses 3", code_name="RAMSES")

        # Test str() conversion
        assert str(ramses) == "[Ramses 3] post-processing code"


class TestSimulationCode(object):
    def test_simu_code_init(self):
        """
        Tests SimulationCode instance initialisation
        """
        ramses = SimulationCode(name="Ramses 3", code_name="RAMSES")

        # Test str() conversion
        assert str(ramses) == "[Ramses 3] simulation code"

    def test_simucode_equality(self):
        """
        Tests rich comparison method SimulationCode.__eq__()
        """
        ramses = SimulationCode(name="Ramses 3 (MHD)", code_name="Ramses", code_version="3.10.1", alias="RAMSES_3",
                                url="http://www.my-code.net", description="This is a fair description")

        r2 = SimulationCode(name=ramses.name, code_version=ramses.code_version, uid=ramses.uid,
                            code_name=ramses.code_name, alias=ramses.alias, url=ramses.url,
                            description=ramses.description)

        assert r2 == ramses

        # Add physical processes
        pp1 = ramses.physical_processes.add(PhysicalProcess(physics=Physics.StarFormation, description="descr sf"))
        pp2 = ramses.physical_processes.add(PhysicalProcess(physics=Physics.Hydrodynamics, description="descr hydro"))
        pp3 = ramses.physical_processes.add(PhysicalProcess(physics=Physics.SelfGravity, description="descr self G"))
        pp4 = ramses.physical_processes.add(PhysicalProcess(physics=Physics.SupernovaeFeedback, description="SN feedback"))

        # Protocol physical process list differ => not equals
        r2.physical_processes.add(pp1)
        r2.physical_processes.add(pp2)
        r2.physical_processes.add(pp3)
        assert r2 != ramses

        # Same physical processes => identical simulation codes
        r2.physical_processes.add(pp4)
        assert r2 == ramses

    def test_simucode_hdf5_io(self, tmp_path):
        """
        Tests saving/loading Protocol from HDF5 dict

        Parameters
        ----------
        tmp_path: temporary path  PyTest fixture
        """
        ramses = SimulationCode(name="Ramses 3 (MHD)", code_name="Ramses", code_version="3.10.1", alias="RAMSES_3",
                                url="http://www.my-code.net", description="This is a fair description")

        ramses.algorithms.add(Algorithm(algo_type=AlgoType.AdaptiveMeshRefinement, description="AMR descr"))
        ramses.algorithms.add(Algorithm(algo_type=AlgoType.Godunov, description="Godunov scheme"))
        ramses.algorithms.add(Algorithm(algo_type=AlgoType.HLLCRiemann, description="HLLC Riemann solver"))
        ramses.algorithms.add(Algorithm(algo_type=AlgoType.PoissonMultigrid, description="Multigrid Poisson solver"))
        ramses.algorithms.add(Algorithm(algo_type=AlgoType.ParticleMesh, description="PM solver"))

        # Add input parameters
        ramses.input_parameters.add(InputParameter(key="levelmin", name="levelmin",
                                                   description="min. level of AMR refinement"))
        ramses.input_parameters.add(InputParameter(key="levelmax", name="levelmax",
                                                   description="max. level of AMR refinement"))

        # Add physical processes
        ramses.physical_processes.add(PhysicalProcess(physics=Physics.StarFormation, description="descr sf"))
        ramses.physical_processes.add(PhysicalProcess(physics=Physics.Hydrodynamics, description="descr hydro"))
        ramses.physical_processes.add(PhysicalProcess(physics=Physics.SelfGravity, description="descr self G"))
        ramses.physical_processes.add(PhysicalProcess(physics=Physics.SupernovaeFeedback, description="SN feedback"))

        # Dummy project
        proj = Project(category=ProjectCategory.GalacticDynamics, project_title="My project")
        proj.simulations.add(Simulation(ramses, name="My simu énorme"))

        # Save study
        fname = str(tmp_path / "study.h5")
        study = SimulationStudy(project=proj)
        study.save_HDF5(fname)

        # Reload study
        study_loaded = SimulationStudy.load_HDF5(fname)
        proj_loaded = study_loaded.project
        simu_loaded = proj_loaded.simulations["My simu énorme"]
        ramses_loaded = simu_loaded.simulation_code
        assert ramses_loaded == ramses


__all__ = ["TestAlgoType", "TestPhysics", "TestAlgorithm", "TestInputParameter", "TestSimulationCode"]





# -*- coding: utf-8 -*-
# -*- coding: utf-8 -*-
# This file is part of the 'astrophysix' Python package.
#
# Copyright © Commissariat a l'Energie Atomique et aux Energies Alternatives (CEA)
#
#  FREE SOFTWARE LICENCING
#  -----------------------
# This software is governed by the CeCILL license under French law and abiding by the rules of distribution of free
# software. You can use, modify and/or redistribute the software under the terms of the CeCILL license as circulated by
# CEA, CNRS and INRIA at the following URL: "http://www.cecill.info". As a counterpart to the access to the source code
# and rights to copy, modify and redistribute granted by the license, users are provided only with a limited warranty
# and the software's author, the holder of the economic rights, and the successive licensors have only limited
# liability. In this respect, the user's attention is drawn to the risks associated with loading, using, modifying
# and/or developing or reproducing the software by the user in light of its specific status of free software, that may
# mean that it is complicated to manipulate, and that also therefore means that it is reserved for developers and
# experienced professionals having in-depth computer knowledge. Users are therefore encouraged to load and test the
# software's suitability as regards their requirements in conditions enabling the security of their systems and/or data
# to be ensured and, more generally, to use and operate it in the same conditions as regards security. The fact that
# you are presently reading this means that you have had knowledge of the CeCILL license and that you accept its terms.
#
#
# COMMERCIAL SOFTWARE LICENCING
# -----------------------------
# You can obtain this software from CEA under other licencing terms for commercial purposes. For this you will need to
# negotiate a specific contract with a legal representative of CEA.
#
from __future__ import print_function, unicode_literals, division, absolute_import  # Python 2 and 3 compatibility
from future.builtins import str
import sys
import logging
import datetime
if sys.version_info.major == 2:
    import pytz

log = logging.getLogger("astrophysix.simdm")


class DatetimeUtil(object):
    @staticmethod
    def utc_now():
        """Returns a UTC timezone-aware datetime corresponding to now"""
        if sys.version_info.major > 2:  # Python 3.+
            return datetime.datetime.now(datetime.timezone.utc)
        else:  # Python 2
            return datetime.datetime.utcnow().replace(tzinfo=pytz.UTC)

    @staticmethod
    def today():
        return datetime.datetime.today()

    @staticmethod
    def utc_from_timestamp(tms):
        """Get UTC timezone-aware datetime corresponding to a given timestamp"""
        if sys.version_info.major > 2:  # Python 3.+
            return datetime.datetime.utcfromtimestamp(tms).replace(tzinfo=datetime.timezone.utc)
        else:  # Python 2
            return datetime.datetime.utcfromtimestamp(tms).replace(tzinfo=pytz.UTC)

    @staticmethod
    def utc_to_timestamp(dt):
        """
        Get the POSIX timestamp corresponding to a datetime instance
        see also : https://docs.python.org/3.3/library/datetime.html#datetime.datetime.timestamp

        Parameters
        ----------
        dt: ``datetime.datetime``
            UTC timezone-aware datetime object to convert

        Returns
        -------
        ts: ``float``
            POSIX timestamp corresponding to dt
        """
        if sys.version_info.major > 3 or (sys.version_info.major == 3 and sys.version_info.minor >= 3):  # Python 3.3+
            return dt.timestamp()
        else:
            ts = (dt - datetime.datetime(1970, 1, 1, tzinfo=pytz.UTC)).total_seconds()
            return ts

    @staticmethod
    def utc_from_string(s, fmt):
        """Returns a UTC timezone-aware datetime corresponding to a given string and a given format"""
        if sys.version_info.major > 2:  # Python 3.+
            return datetime.datetime.strptime(s, fmt).replace(tzinfo=datetime.timezone.utc)
        else:  # Python 2
            return datetime.datetime.strptime(s, fmt).replace(tzinfo=pytz.UTC)


__all__ = ["DatetimeUtil"]



# -*- coding: utf-8 -*-
# This file is part of the 'astrophysix' Python package.
#
# Copyright © Commissariat a l'Energie Atomique et aux Energies Alternatives (CEA)
#
#  FREE SOFTWARE LICENCING
#  -----------------------
# This software is governed by the CeCILL license under French law and abiding by the rules of distribution of free
# software. You can use, modify and/or redistribute the software under the terms of the CeCILL license as circulated by
# CEA, CNRS and INRIA at the following URL: "http://www.cecill.info". As a counterpart to the access to the source code
# and rights to copy, modify and redistribute granted by the license, users are provided only with a limited warranty
# and the software's author, the holder of the economic rights, and the successive licensors have only limited
# liability. In this respect, the user's attention is drawn to the risks associated with loading, using, modifying
# and/or developing or reproducing the software by the user in light of its specific status of free software, that may
# mean that it is complicated to manipulate, and that also therefore means that it is reserved for developers and
# experienced professionals having in-depth computer knowledge. Users are therefore encouraged to load and test the
# software's suitability as regards their requirements in conditions enabling the security of their systems and/or data
# to be ensured and, more generally, to use and operate it in the same conditions as regards security. The fact that
# you are presently reading this means that you have had knowledge of the CeCILL license and that you accept its terms.
#
#
# COMMERCIAL SOFTWARE LICENCING
# -----------------------------
# You can obtain this software from CEA under other licencing terms for commercial purposes. For this you will need to
# negotiate a specific contract with a legal representative of CEA.
#
import sys

from future.builtins import str


class Stringifiable(object):
    if sys.version_info.major > 2:  # Python 3.+
        def __str__(self):
            return self.__unicode__()
    else:  # Python 2
        def __str__(self):
            return self.__unicode__().encode('utf8')

    def __unicode__(self):
        """
        String representation of the instance
        """
        raise NotImplementedError()

    def __repr__(self):
        return self.__str__()

    @staticmethod
    def is_type_string(s):
        """
        If the object is not a string object (basestring in Python 2, str in Python 3), returns False. Otherwise returns
        True.

        Parameters
        ----------
        s: object to test.
        """
        if sys.version_info.major > 2:  # Python 3.+
            return isinstance(s, str)
        else:  # Python 2
            return isinstance(s, basestring)

    @staticmethod
    def cast_string(s, valid_empty=True):
        """
        Tries to cast an object into a Python 3 compatible unicode string. If the object is not a string object
        (basestring in Python 2, str in Python 3), or if the string is empty and 'valid_empty' is False, raises a
        TypeError.

        Parameters
        ----------
        s: object to cast into a Python 3 compatible string
        valid_empty: Are empty strings valid ? Default true.

        Returns
        -------
        p3s: Python 3 compatible unicode str
        """
        if sys.version_info.major > 2:  # Python 3.+
            if not isinstance(s, str) or (not valid_empty and len(s) == 0):
                raise TypeError("Invalid string")
            return s
        else:  # Python 2
            if not isinstance(s, basestring) or (not valid_empty and len(s) == 0):
                raise TypeError("Invalid string")
            # Cast to Python 3 unicode str
            return str(s)


__all__ = ["Stringifiable"]



# -*- coding: utf-8 -*-
# This file is part of the 'astrophysix' Python package.
#
# Copyright © Commissariat a l'Energie Atomique et aux Energies Alternatives (CEA)
#
#  FREE SOFTWARE LICENCING
#  -----------------------
# This software is governed by the CeCILL license under French law and abiding by the rules of distribution of free
# software. You can use, modify and/or redistribute the software under the terms of the CeCILL license as circulated by
# CEA, CNRS and INRIA at the following URL: "http://www.cecill.info". As a counterpart to the access to the source code
# and rights to copy, modify and redistribute granted by the license, users are provided only with a limited warranty
# and the software's author, the holder of the economic rights, and the successive licensors have only limited
# liability. In this respect, the user's attention is drawn to the risks associated with loading, using, modifying
# and/or developing or reproducing the software by the user in light of its specific status of free software, that may
# mean that it is complicated to manipulate, and that also therefore means that it is reserved for developers and
# experienced professionals having in-depth computer knowledge. Users are therefore encouraged to load and test the
# software's suitability as regards their requirements in conditions enabling the security of their systems and/or data
# to be ensured and, more generally, to use and operate it in the same conditions as regards security. The fact that
# you are presently reading this means that you have had knowledge of the CeCILL license and that you accept its terms.
#
#
# COMMERCIAL SOFTWARE LICENCING
# -----------------------------
# You can obtain this software from CEA under other licencing terms for commercial purposes. For this you will need to
# negotiate a specific contract with a legal representative of CEA.
#
from __future__ import print_function, unicode_literals, division, absolute_import  # Python 2 and 3 compatibility

import logging

from future.builtins import str, int
import pytest

from astrophysix import units as U
from astrophysix.simdm import Project, ProjectCategory, SimulationStudy
from astrophysix.simdm.experiment import Simulation
from astrophysix.simdm.protocol import SimulationCode
from astrophysix.simdm.results import GenericResult
from astrophysix.simdm.catalogs import TargetObject, ObjectProperty, ObjectPropertyGroup, Catalog, PropertySortFlag, \
    PropertyFilterFlag
from astrophysix.simdm.utils import DataType


class TestObjectProperty(object):
    def test_target_object_property_init(self):
        """
        Tests target object property instance initialisation
        """
        # Object property initialisation
        x = ObjectProperty(property_name="pos_x", description="My property descr.", unit=U.Mpc, dtype=DataType.INTEGER,
                           filter_flag=PropertyFilterFlag.BASIC_FILTER, sort_flag=PropertySortFlag.ADVANCED_SORT)

        assert x.property_name == "pos_x"
        assert x.description == "My property descr."
        assert x.unit == U.Mpc
        assert x.filter_flag == PropertyFilterFlag.BASIC_FILTER
        assert x.sort_flag == PropertySortFlag.ADVANCED_SORT
        assert x.datatype == DataType.INTEGER

        # Test str() conversion
        assert str(x) == "'pos_x' target object property"

    def test_setting_object_property_name_exc(self):
        """
        Tests that a target object  property defined without name raises an exception
        """
        # Tests target object property defined without name
        with pytest.raises(AttributeError, match="ObjectProperty 'property_name' attribute is not defined \\(mandatory\\)."):
            objprop_exc = ObjectProperty()  # No object property name defined

        x = ObjectProperty(property_name="pos_x")

        # -------------- Tests invalid object property name setting exception -------------------- #
        empty_name_err = "ObjectProperty 'property_name' property is not a valid \\(non-empty\\) string."
        with pytest.raises(AttributeError, match=empty_name_err):
            x.property_name = -1

        with pytest.raises(AttributeError, match=empty_name_err):
            x.property_name = ""
        # ---------------------------------------------------------------------------------------- #

        # Valid => should not raise any exception
        x.property_name = "Disk-like galaxy"

    def test_setting_object_property_description(self):
        """
        Tests setting object property description property
        """
        vx = ObjectProperty(property_name="v_x")

        # ---------------------- Tests invalid object property description setting exception ------------------------- #
        with pytest.raises(AttributeError, match="ObjectProperty 'description' property is not a valid string."):
            vx.description = {'a': [0, 1, 3]}

        # Valid => should not raise any exception
        vx.description = "My target object velocity property"
        # ------------------------------------------------------------------------------------------------------------ #

    def test_setting_object_property_sort_filter_flags(self):
        """
        Tests setting object property filter/sort flag property
        """
        vx = ObjectProperty(property_name="v_x")

        # ----------------------- Tests invalid object property filter flag setting exception ------------------------ #
        with pytest.raises(AttributeError, match="Object property 'filter_flag' property error : No PropertyFilterFlag "
                                                 "defined with the flag 'UNKNOWN_FLAG'."):
            vx.filter_flag = "UNKNOWN_FLAG"

        with pytest.raises(AttributeError, match="Object property 'filter_flag' attribute is not a valid "
                                                 "PropertyFilterFlag enum value."):
            vx.filter_flag = {"a": 0.45}
        # ------------------------------------------------------------------------------------------------------------ #

        # ------------------------- Tests invalid object property sort flag setting exception ------------------------ #
        with pytest.raises(AttributeError, match="Object property 'sort_flag' property error : No PropertySortFlag "
                                                 "defined with the flag 'UNKNOWN_SORT'."):
            vx.sort_flag = "UNKNOWN_SORT"

        with pytest.raises(AttributeError, match="Object property 'sort_flag' attribute is not a valid PropertySortFlag"
                                                 " enum value."):
            vx.sort_flag = {"a": 0.45}
        # ------------------------------------------------------------------------------------------------------------ #

        # Valid => should not raise any exception
        vx.filter_flag = PropertyFilterFlag.BASIC_FILTER
        vx.sort_flag = PropertySortFlag.ADVANCED_SORT
        assert vx.filter_flag == PropertyFilterFlag.BASIC_FILTER
        assert vx.sort_flag == PropertySortFlag.ADVANCED_SORT

    def test_setting_object_property_datatype(self):
        """
        Tests setting object property datatype property
        """
        vx = ObjectProperty(property_name="v_x", dtype=DataType.INTEGER)
        assert vx.datatype == DataType.INTEGER

        # ----------------------- Tests invalid object property datatype setting exception ------------------------ #
        with pytest.raises(AttributeError, match="Object property 'datatype' error : No DataType "
                                                 "defined with the key 'unkn_dtype'."):
            vx.datatype = "unkn_dtype"

        with pytest.raises(AttributeError, match="Object property 'datatype' attribute is not a valid "
                                                 "DataType enum value."):
            vx.datatype = {"bcde": 1.5, 23: [1.2, 3.4, 5.6]}
        # ------------------------------------------------------------------------------------------------------------ #

        # Valid => should not raise any exception
        vx.datatype = DataType.BOOLEAN
        assert vx.datatype == DataType.BOOLEAN

    def test_setting_object_property_unit(self):
        """
        Tests setting object property unit property
        """
        alpha = ObjectProperty(property_name="alpha")

        # --------------------------- Tests invalid object property unit setting exception --------------------------- #
        with pytest.raises(AttributeError, match="Object property 'unit' property error : Unknown unit name 'fairies'."):
            alpha.unit = "fairies"

        invalid_objprop_unit_err = "Object property 'unit' property is not a valid \\(non-empty\\) string."
        with pytest.raises(AttributeError, match=invalid_objprop_unit_err):
            alpha.unit = ""

        with pytest.raises(AttributeError, match=invalid_objprop_unit_err):
            alpha.unit = ["kg", "V", "cm", "Gauss"]
        # ------------------------------------------------------------------------------------------------------------ #

        # Valid => should not raise any exception
        alpha.unit = U.mGauss
        assert alpha.unit == U.mGauss

    def test_object_property_display_name(self):
        """Tests object property display name name + (unit)"""
        vx = ObjectProperty(property_name="V_x", unit=U.km_s)
        assert vx.display_name == "V_x (\\textrm{km}\\cdot\\textrm{s}^{-1})"

    def test_object_property_equality(self):
        """
        Tests object property rich comparison method ObjectProperty.__eq__()
        """
        vx = ObjectProperty(property_name="v_x", sort_flag=PropertySortFlag.BASIC_SORT, unit=U.km_s,
                            filter_flag=PropertyFilterFlag.BASIC_FILTER, dtype=DataType.RATIONAL,
                            description="Galaxy global velocity coordinate along x-axis")

        # Different UUID => not equals
        assert vx != ObjectProperty(property_name=vx.property_name, description=vx.description, sort_flag=vx.sort_flag,
                                    filter_flag=vx.filter_flag, unit=vx.unit)

        # Different name => not equals
        vx2 = ObjectProperty(property_name="different name", description=vx.description, sort_flag=vx.sort_flag,
                             filter_flag=vx.filter_flag, unit=vx.unit, uid=vx.uid)
        assert vx != vx2
        vx2.property_name = vx.property_name

        # Different description => not equals
        vx2.description = "Different description"
        assert vx != vx2
        vx2.description = vx.description

        # Different filter/sort flags => not equals
        vx2.filter_flag = PropertyFilterFlag.NO_FILTER
        assert vx != vx2
        vx2.filter_flag = vx.filter_flag
        vx2.sort_flag = PropertySortFlag.NO_SORT
        assert vx != vx2
        vx2.sort_flag = vx.sort_flag

        # Different datatpy => not equals
        vx2.datatype = DataType.INTEGER
        assert vx != vx2
        vx2.datatype = vx.datatype

        # Different unit => not equals
        vx2.unit = U.pc
        assert vx != vx2
        vx2.unit = vx.unit

        # Identical object properties
        assert vx == vx2

    def test_target_object_property_galactica_validity_checks(self, caplog):
        """
        Tests target object property Galactica validity checks

        Parameters
        ----------
        caplog: captured log PyTest fixture
        """
        # Create target object
        tobj = TargetObject(name="Spiral galaxy")
        tobj.object_properties.add(ObjectProperty(property_name="This is a way too long property name for a 'Spiral "
                                                                "galaxy' target object"))

        # Object property name too long
        tobj.galactica_validity_check()
        assert caplog.record_tuples[-1] == ('astrophysix.simdm', logging.WARNING,
                                            "'This is a way too long property name for a 'Spiral galaxy' target object'"
                                            " target object property name is too long for Galactica (max. 64 "
                                            "characters).")

    def test_object_property_hdf5_io(self, tmp_path):
        """
        Tests saving/loading ObjectProperty from HDF5 file

        Parameters
        ----------
        tmp_path: temporary path  PyTest fixture
        """
        res = GenericResult(name="Key result 1 !")
        tobj = TargetObject(name="Spiral galaxy")

        # Add properties + property groups
        vx = tobj.object_properties.add(ObjectProperty(property_name="v_x", sort_flag=PropertySortFlag.BASIC_SORT,
                                                       unit=U.km_s, filter_flag=PropertyFilterFlag.BASIC_FILTER,
                                                       dtype=DataType.RATIONAL,
                                                       description="Galaxy global velocity coordinate along x-axis"))
        cat = res.catalogs.add(Catalog(target_object=tobj, name="Spiral galaxy catalog"))
        ramses = SimulationCode(name="Ramses 3 (MHD)", code_name="Ramses")
        simu = Simulation(simu_code=ramses, name="My simu énorme")
        simu.generic_results.add(res)

        # Dummy project
        proj = Project(category=ProjectCategory.GalacticDynamics, project_title="My project")
        proj.simulations.add(simu)

        # Save study
        fname = str(tmp_path / "study.h5")
        study = SimulationStudy(project=proj)
        study.save_HDF5(fname)

        # Reload study
        study_loaded = SimulationStudy.load_HDF5(fname)
        simu_loaded = study_loaded.project.simulations[simu.name]

        # Compare object properties
        res_loaded = simu_loaded.generic_results[res.name]
        cat_loaded = res_loaded.catalogs[cat.name]
        loaded_vx = cat_loaded.target_object.object_properties[vx.property_name]
        assert vx == loaded_vx


class TestObjectPropertyGroup(object):
    def test_target_object_property_group_init(self):
        """
        Tests target object property group instance initialisation
        """
        # Object property group initialisation
        g = ObjectPropertyGroup(group_name="velocity", description="My group descr.")

        assert g.group_name == "velocity"
        assert g.description == "My group descr."

        # Test str() conversion
        assert str(g) == "'velocity' property group"

    def test_setting_object_property_group_name_exc(self):
        """
        Tests that a target object property group defined without name raises an exception
        """
        # Tests target object property group defined without name
        with pytest.raises(AttributeError, match="ObjectPropertyGroup 'group_name' attribute is not defined \\(mandatory\\)."):
            pgroup_exc = ObjectPropertyGroup()  # No object property group name defined

        g = ObjectPropertyGroup(group_name="velocity")

        # ------------ Tests invalid object property group name setting exception ---------------- #
        empty_name_err = "ObjectPropertyGroup 'group_name' property is not a valid \\(non-empty\\) string."
        with pytest.raises(AttributeError, match=empty_name_err):
            g.group_name = -1

        with pytest.raises(AttributeError, match=empty_name_err):
            g.group_name = ""
        # ---------------------------------------------------------------------------------------- #

        # Valid => should not raise any exception
        g.group_name = "Velocity"

    def test_setting_object_property_group_description(self):
        """
        Tests setting object property description property
        """
        g = ObjectPropertyGroup(group_name="velocities")

        # ------------------- Tests invalid object property group description setting exception ---------------------- #
        with pytest.raises(AttributeError, match="ObjectPropertyGroup 'description' property is not a valid string."):
            g.description = {'a': [0, 1, 3]}

        # Valid => should not raise any exception
        g.description = "My target object velocity property group"
        # ------------------------------------------------------------------------------------------------------------ #

    def test_object_property_group_equality(self):
        """
        Tests object property rich comparison method ObjectProperty.__eq__()
        """
        vel = ObjectPropertyGroup(group_name="velocity", description="My descr.")

        # Different UUID => not equals
        assert vel != ObjectPropertyGroup(group_name=vel.group_name, description=vel.description)

        # Different name => not equals
        vel2 = ObjectPropertyGroup(group_name="other name", description=vel.description, uid=vel.uid)
        assert vel != vel2
        vel2.group_name = vel.group_name
        assert vel == vel2

        # Different description => not equals
        vel2.description = "Different description"
        assert vel != vel2
        vel2.description = vel.description

        # Group property list differ => not equals
        vx = vel.group_properties.add(ObjectProperty(property_name="vel_x"))
        vy = vel.group_properties.add(ObjectProperty(property_name="vel_y"))
        vel2.group_properties.add(vx)
        assert vel2 != vel
        vel2.group_properties.add(vy)

        # Identical object property groups
        assert vel == vel2

    def test_target_object_property_group_galactica_validity_checks(self, caplog):
        """
        Tests target object property group Galactica validity checks

        Parameters
        ----------
        caplog: captured log PyTest fixture
        """
        # Create target object with object property group
        tobj = TargetObject(name="Spiral galaxy")
        tobj.property_groups.add(ObjectPropertyGroup(group_name="This is a way too long object property group name"))

        # Object property name too long
        tobj.galactica_validity_check()
        assert caplog.record_tuples[-1] == ('astrophysix.simdm', logging.WARNING,
                                            "'This is a way too long object property group name' property group name "
                                            "is too long for Galactica (max. 32 characters).")

    def test_object_property_group_hdf5_io(self, tmp_path):
        """
        Tests saving/loading ObjectPropertyGroup from HDF5 file

        Parameters
        ----------
        tmp_path: temporary path  PyTest fixture
        """
        res = GenericResult(name="Key result 1 !")
        tobj = TargetObject(name="Spiral galaxy")

        # Add properties + property groups
        vx = tobj.object_properties.add(ObjectProperty(property_name="v_x"))
        vy = tobj.object_properties.add(ObjectProperty(property_name="v_y"))
        vz = tobj.object_properties.add(ObjectProperty(property_name="v_z"))
        vel = tobj.property_groups.add(ObjectPropertyGroup(group_name="velocity", description="Galaxy global velocity"))
        vel.group_properties.add(vx)
        vel.group_properties.add(vy)
        vel.group_properties.add(vz)
        cat = res.catalogs.add(Catalog(target_object=tobj, name="Spiral galaxy catalog"))
        ramses = SimulationCode(name="Ramses 3 (MHD)", code_name="Ramses")
        simu = Simulation(simu_code=ramses, name="My simu énorme")
        simu.generic_results.add(res)

        # Dummy project
        proj = Project(category=ProjectCategory.GalacticDynamics, project_title="My project")
        proj.simulations.add(simu)

        # Save study
        fname = str(tmp_path / "study.h5")
        study = SimulationStudy(project=proj)
        study.save_HDF5(fname)

        # Reload study
        study_loaded = SimulationStudy.load_HDF5(fname)
        simu_loaded = study_loaded.project.simulations[simu.name]

        # Compare object propertiy groups
        res_loaded = simu_loaded.generic_results[res.name]
        cat_loaded = res_loaded.catalogs[cat.name]
        tobj_loaded = cat_loaded.target_object
        velocities_loaded = tobj_loaded.property_groups[vel.group_name]
        assert vel == velocities_loaded


class TestTargetObject(object):
    def test_target_object_init(self):
        """
        Tests target object instance initialisation
        """
        # Target object initialisation
        tobj = TargetObject(name="Spiral galaxy", description="This is a disk-like galaxy")
        tobj.object_properties.add(ObjectProperty(property_name="pos_x"))
        tobj.object_properties.add(ObjectProperty(property_name="pos_y"))
        tobj.object_properties.add(ObjectProperty(property_name="mass"))

        assert tobj.name == "Spiral galaxy"
        assert tobj.description == "This is a disk-like galaxy"
        assert len(tobj.object_properties) == 3

        # Test str() conversion
        assert str(tobj) == "'Spiral galaxy' target object"

    def test_target_object_property_insertion(self):
        """
        Tests exceptions raised upon object property(-group) insertion into TargetObject instances
        """
        tobj = TargetObject(name="Spiral galaxy", description="This is a disk-like galaxy")
        alpha = ObjectProperty(property_name='alpha')
        delta = ObjectProperty(property_name='delta')
        beta = ObjectProperty(property_name='beta')
        gamma = ObjectProperty(property_name='gamma')

        # Tries to add property to a group and then add group before the property has been added to the TargetObject
        # property list => raises an error
        g = ObjectPropertyGroup(group_name="G1")
        g.group_properties.add(alpha)
        g.group_properties.add(delta)
        with pytest.raises(AttributeError, match="{p!s} does not belong to this TargetObject object property "
                                                 "list.".format(p=alpha)):
            tobj.property_groups.add(g)

        # Tries to add a property to an already registered group in the TargetObject property group list while the
        # property has not added yet to the TargetObject property list => raises an error
        tobj.object_properties.add(alpha)
        tobj.object_properties.add(delta)
        tobj.property_groups.add(g)  # => Ok
        with pytest.raises(AttributeError, match="{p!s} does not belong to this TargetObject object property "
                                                 "list.".format(p=beta)):
            g.group_properties.add(beta)

        # Valid => no error
        tobj.object_properties.add(gamma)
        g.group_properties.add(gamma)  # Ok

    def test_deleting_object_prop_from_targobj_while_inserted_in_group(self):
        """
        Tests object property deletion from TargetObject's property list while object property is included in one of the
        TargetObject's ObjectPropertyGroup.
        """
        tobj = TargetObject(name="Pre-stellar core")
        x = tobj.object_properties.add(ObjectProperty(property_name="pos_x"))
        y = tobj.object_properties.add(ObjectProperty(property_name="pos_y"))
        z = tobj.object_properties.add(ObjectProperty(property_name="pos_z"))
        pos = tobj.property_groups.add(ObjectPropertyGroup(group_name="position"))
        pos.group_properties.add(x)
        pos.group_properties.add(y)
        pos.group_properties.add(z)

        # Tries to delete x, while it is in 'position' group
        with pytest.raises(AttributeError, match="''pos_x' target object property' cannot be deleted, the following "
                                                 "items depend on it \\(try to delete them first\\) : \\['Pre-stellar "
                                                 "core' target object - 'position' property group - 'pos_x' target "
                                                 "object property\\]."):
            del tobj.object_properties[x.property_name]

        # Delete property from 'position' group first
        del pos.group_properties[x.property_name]

        # THEN delete Property from TargetObject => Ok
        del tobj.object_properties[x.property_name]

    def test_setting_target_object_name_exc(self):
        """
        Tests that a target object defined without name raises an exception
        """
        # Tests target object defined without name
        with pytest.raises(AttributeError, match="TargetObject 'name' attribute is not defined \\(mandatory\\).") as e:
            tobj_exc = TargetObject()  # No target object name defined

        tobj = TargetObject(name="Spiral galaxy")

        # --------------- Tests invalid target object name setting exception --------------------- #
        empty_name_err = "TargetObject 'name' property is not a valid \\(non-empty\\) string."
        with pytest.raises(AttributeError, match=empty_name_err):
            tobj.name = -1

        with pytest.raises(AttributeError, match=empty_name_err):
            tobj.name = ""
        # ---------------------------------------------------------------------------------------- #

        # Valid => should not raise any exception
        tobj.name = "Disk-like galaxy"

    def test_setting_target_object_description(self):
        """
        Tests setting target object description property
        """
        tobj = TargetObject(name="Pre-stellar core")

        # ------------------------ Tests invalid target object description setting exception ------------------------- #
        with pytest.raises(AttributeError, match="TargetObject 'description' property is not a valid string."):
            tobj.description = 0

        # Valid => should not raise any exception
        tobj.description = "My target object"
        # ------------------------------------------------------------------------------------------------------------ #

    def test_targetobject_equality(self):
        """
        Tests target object rich comparison method Catalog.__eq__()
        """
        tobj = TargetObject(name="Elliptic galaxy", description="Reddish roundish old galaxy")

        # Different UUID => not equals
        assert tobj != TargetObject(name=tobj.name, description=tobj.description)

        # Different name => not equals
        assert tobj != TargetObject(name="Name", description=tobj.description, uid=tobj.uid)

        # Different description => not equals
        assert tobj != TargetObject(name=tobj.name, description="Alt. description", uid=tobj.uid)

        # Identical target objects
        t = TargetObject(name=tobj.name, description=tobj.description, uid=tobj.uid)
        assert tobj == t

        # Add object properties
        a1 = tobj.object_properties.add(ObjectProperty(property_name="property #1"))
        a2 = tobj.object_properties.add(ObjectProperty(property_name="property #2"))
        # Object property list differ => not equals
        t.object_properties.add(a1)
        assert t != tobj
        # Identical object properties
        t.object_properties.add(a2)
        assert t == tobj

        # Add Object property groups
        g = tobj.property_groups.add(ObjectPropertyGroup(group_name="Group #1"))
        g.group_properties.add(a1)
        g.group_properties.add(a2)
        # Object property group list differ => not equals
        assert t != tobj
        # Identical object properties
        t.property_groups.add(g)
        assert t == tobj

    def test_target_object_galactica_validity_checks(self, caplog):
        """
        Tests target object Galactica validity checks

        Parameters
        ----------
        caplog: captured log PyTest fixture
        """
        # Create target object
        tobj = TargetObject(name="Spiral galaxy")

        # Target object name too long
        tobj.name = "Spiral galaxy with boxy shapes but not strictly speaking disky"
        tobj.galactica_validity_check()
        assert caplog.record_tuples[-1] == ('astrophysix.simdm', logging.WARNING,
                                            "'Spiral galaxy with boxy shapes but not strictly speaking disky' target "
                                            "object name is too long for Galactica (max. 32 characters).")

    def test_target_object_hdf5_io(self, tmp_path):
        """
        Tests saving/loading TargetObject from HDF5 file

        Parameters
        ----------
        tmp_path: temporary path  PyTest fixture
        """
        res = GenericResult(name="Key result 1 !")
        tobj = TargetObject(name="Spiral galaxy", description="My object full description")

        # Add properties + property groups
        alpha = tobj.object_properties.add(ObjectProperty(property_name='alpha'))
        beta = tobj.object_properties.add(ObjectProperty(property_name='beta'))
        g = tobj.property_groups.add(ObjectPropertyGroup(group_name="G"))
        g.group_properties.add(alpha)
        g.group_properties.add(beta)

        cat = res.catalogs.add(Catalog(target_object=tobj, name="Spiral galaxy catalog"))
        ramses = SimulationCode(name="Ramses 3 (MHD)", code_name="Ramses")
        simu = Simulation(simu_code=ramses, name="My simu énorme")
        simu.generic_results.add(res)

        # Dummy project
        proj = Project(category=ProjectCategory.GalacticDynamics, project_title="My project")
        proj.simulations.add(simu)

        # Save study
        fname = str(tmp_path / "study.h5")
        study = SimulationStudy(project=proj)
        study.save_HDF5(fname)

        # Reload study
        study_loaded = SimulationStudy.load_HDF5(fname)
        simu_loaded = study_loaded.project.simulations[simu.name]

        # Compare target objects
        res_loaded = simu_loaded.generic_results[res.name]
        cat_loaded = res_loaded.catalogs[cat.name]
        assert tobj == cat_loaded.target_object


__all__ = ["TestTargetObject"]



# -*- coding: utf-8 -*-
# This file is part of the 'astrophysix' Python package.
#
# Copyright © Commissariat a l'Energie Atomique et aux Energies Alternatives (CEA)
#
#  FREE SOFTWARE LICENCING
#  -----------------------
# This software is governed by the CeCILL license under French law and abiding by the rules of distribution of free
# software. You can use, modify and/or redistribute the software under the terms of the CeCILL license as circulated by
# CEA, CNRS and INRIA at the following URL: "http://www.cecill.info". As a counterpart to the access to the source code
# and rights to copy, modify and redistribute granted by the license, users are provided only with a limited warranty
# and the software's author, the holder of the economic rights, and the successive licensors have only limited
# liability. In this respect, the user's attention is drawn to the risks associated with loading, using, modifying
# and/or developing or reproducing the software by the user in light of its specific status of free software, that may
# mean that it is complicated to manipulate, and that also therefore means that it is reserved for developers and
# experienced professionals having in-depth computer knowledge. Users are therefore encouraged to load and test the
# software's suitability as regards their requirements in conditions enabling the security of their systems and/or data
# to be ensured and, more generally, to use and operate it in the same conditions as regards security. The fact that
# you are presently reading this means that you have had knowledge of the CeCILL license and that you accept its terms.
#
#
# COMMERCIAL SOFTWARE LICENCING
# -----------------------------
# You can obtain this software from CEA under other licencing terms for commercial purposes. For this you will need to
# negotiate a specific contract with a legal representative of CEA.
#
from __future__ import unicode_literals

from future.builtins import str
from enum import Enum
import hashlib
import re
import os
import tempfile

from .strings import Stringifiable
from .dt_utils import DatetimeUtil


class FileType(Enum):
    """File type enum

    Example
    -------
    >>> ft = FileType.ASCII_FILE
    >>> ft.alias
    "ASCII"
    >>> ft.extension_list
    [".dat", ".DAT", ".txt", ".TXT", ".ini", ".INI"]
    """
    PNG_FILE = ("PNG", [".png", ".PNG"])
    JPEG_FILE = ("JPEG", [".jpg", ".jpeg", ".JPG", ".JPEG"])
    FITS_FILE = ("FITS", [".fits", ".FITS"])
    TARGZ_FILE = ("TARGZ", [".tar.gz", ".TAR.GZ", ".TAR.gz", ".tar.GZ", ".tgz", ".TGZ"])
    PICKLE_FILE = ("PICKLE", [".pkl", ".PKL", ".pickle", ".sav", ".save"])
    HDF5_FILE = ("HDF5", [".h5", ".H5", ".hdf5", ".HDF5"])
    JSON_FILE = ("JSON", [".json", ".JSON"])
    CSV_FILE = ("CSV", [".csv", ".CSV"])
    YAML_FILE = ("YAML", [".yml", ".YML", ".yaml", ".YAML"])
    ASCII_FILE = ("ASCII", [".dat", ".DAT", ".txt", ".TXT", ".ini", ".INI"])
    XML_FILE = ("XML", ['.xml', '.XML'])

    def __init__(self, alias, ext_list):
        self._alias = alias
        self._extension_list = ext_list

    @property
    def alias(self):
        """
        Returns file type alias
        """
        return self._alias

    @property
    def extension_list(self):
        """
        Returns file type valid extension list
        """
        return self._extension_list

    @property
    def default_extension(self):
        """Returns the first item in the file type extension list"""
        return self._extension_list[0]

    @property
    def file_regexp(self):
        """
        Returns filename matching regular expression for the current file type
        """
        ext = "|".join([e[1:] for e in self._extension_list])
        return re.compile("(?P<basename>.+)\\.(?P<extension>({ext:s}))$".format(ext=ext))

    @classmethod
    def from_alias(cls, alias):
        """
        Find a FileType according to its alias

        Parameters
        ----------
        alias: :obj:`string`
            required file type alias

        Returns
        -------
        ft: :class:`~astrophysix.utils.file.FileType`
            File type matching the requested alias.

        Raises
        ------
        ValueError
            if requested alias does not match any file type.

        Example
        -------
            >>> ft = FileType.from_alias("PNG")
            >>> ft.extension_list
            [".png", ".PNG"]
            >>> ft2 = FileType.from_alias("MY_UNKNOWN_FILETYPE")
            ValuerError: No FileType defined with the alias 'MY_UNKNOWN_FILETYPE'.
        """
        for ftype in cls:
            if ftype.alias == alias:
                return ftype
        raise ValueError("No FileType defined with the alias '{a:s}'.".format(a=alias))

    def __unicode__(self):
        """
        String representation of the enum value. Returns alias.
        """
        return self._alias


class FileUtil(object):
    """
    File path utility abstract class
    """
    @classmethod
    def get_file_type(cls, filepath):
        """
        Find the file type or return None is no matching file type has been recognized.
        """
        cls._check_filename_is_string(filepath)
        filename = os.path.basename(filepath)

        for ftype in FileType:
            if ftype.file_regexp.match(filename) is not None:
                return ftype
        return None

    @staticmethod
    def valid_filepath(filepath, append_extension=None):
        """
        Build a valid filepath with a default file extension, if missing.

        Parameters
        ----------
        filepath: file path
        append_extension: optional appended extension FileType. Default None (do not append any extension)

        Returns
        -------
        fname:
        """
        FileUtil._check_filename_is_string(filepath)

        # Get file absolute path
        abs_filepath = os.path.abspath(filepath)

        if append_extension is None:
            return abs_filepath
        else:
            # Check that append_extension is a valid FileType enum value
            FileUtil._check_valid_file_type(append_extension)

            if append_extension.file_regexp.match(abs_filepath) is None:  # Missing extension
                # Append default extension to filename
                ext = append_extension.default_extension
                return "{fpath!s}{default_ext!s}".format(fpath=abs_filepath, default_ext=ext)
            else:  # Extension is not missing => return absolute file path
                return abs_filepath

    @staticmethod
    def find_file(basename, file_type):
        """
        Search a file with a given base name and a given file type: checks the existence of all possible file extensions

        Parameters
        ----------
        basename: file basename
        file_type: FileType enum value

        Returns
        -------
        fname: the file name if it actually exists, otherwise returns None.
        """
        FileUtil._check_filename_is_string(basename)
        FileUtil._check_valid_file_type(file_type)

        if file_type.file_regexp.match(basename) is not None:
            if os.path.isfile(basename):
                return basename
        else:
            for ext in file_type.extension_list:
                fname = "{bname!s}{extension!s}".format(bname=basename, extension=ext)
                if os.path.isfile(fname):
                    return fname
        return None

    @staticmethod
    def md5sum(fname):
        """
        Compute the md5 checksum of a given file

        Parameter
        ---------
        fname: `string`
            path of the file
        """
        if not os.path.isfile(fname):
            raise IOError("Cannot process md5sum of the file '{fname!s}': the file does not exist.".format(fname=fname))

        hash_md5 = hashlib.md5()
        # Read file in chunks to compute byte checksum using MD5 algorithm
        with open(fname, "rb") as f:
            for chunk in iter(lambda: f.read(4096), b""):
                hash_md5.update(chunk)
        return hash_md5.hexdigest()

    @staticmethod
    def split_fname_extension(filename, file_type):
        """
        Checks a given filename is a valid file name of type 'file_type' and splits the valid filename into a
        (basename, extension) tuple.

        Parameters
        ----------
        filename: ``string``
            file name
        file_type: ``FileType``
            file type

        Returns
        -------
        t: ``tuple``
            (basename, extension) 2-tuple
        """
        FileUtil._check_filename_is_string(filename)
        FileUtil._check_valid_file_type(file_type)

        fname = os.path.basename(filename)
        m = file_type.file_regexp.match(fname)
        if m is not None:
            # Get basename and extension
            return m.group("basename"), m.group("extension")
        else:
            raise AttributeError("'filename' is not a valid {file_type!s} file name.".format(file_type=str(file_type)))

    @staticmethod
    def new_filepath(filename, append_extension, create_dir=True):
        valid_fname = FileUtil.valid_filepath(filename, append_extension)

        file_dir = os.path.dirname(valid_fname)
        if create_dir and not os.path.isdir(file_dir):  # Create directory if it does not exist
            os.makedirs(file_dir)
        return valid_fname

    @staticmethod
    def new_temp_filepath(prefix):
        # Create a temporary file in /tmp/ directory
        today = DatetimeUtil.today()
        temp_file_prefix = ("Astrophysix_{prefix:s}_{year:04d}.{month:02d}.{day:02d}_"
                            "{hour:02d}-{min:02d}-{sec:02d}").format(year=today.year, month=today.month, day=today.day,
                                                                     hour=today.hour, min=today.minute, sec=today.second,
                                                                     prefix=prefix)
        fd, fpath = tempfile.mkstemp(prefix=temp_file_prefix, suffix=".h5")
        os.close(fd)
        return str(fpath)

    @staticmethod
    def locate_file(path, base_directory=None):
        """
        Try to locate a file given its base path and an optional base directory in which to look for an existing file
        with the same filename in any similar subdirectory tree

        Parameters
        ----------
        path: default filepath
        base_directory: Base directory in which an existing file with identical filename must be searched. If None,
        do not search. Default None.

        Returns
        -------
        out_path: found existing filepath, if found. Otherwise returns None.
        """
        FileUtil._check_filename_is_string(path)

        if base_directory is None:  # Do not search in directory
            if os.path.isfile(path):  # File exists => easy, return file path
                return path

            # File does not exist, raises IOError
            raise IOError("Cannot find file '{fname!s}': the file does not exist.".format(fname=path))

        path, fname = os.path.split(path)  # Split directory and filename  e.g. ("/path/to/my/directory/", "file.png")

        # Simple case : look for existing filename in base directory
        out_path = os.path.join(base_directory, fname)
        if os.path.isfile(out_path):  # Found => return path
            return out_path

        # Full directory tree search
        dirs = [d for d in path.split(os.sep) if len(d) > 0]
        for d in dirs[::-1]:
            # Build a similar subdirectory tree path of the filename
            fname = os.path.join(d, fname)

            # Concatenate with the base_directory
            out_path = os.path.join(base_directory, fname)

            if os.path.isfile(out_path):  # Found existing filename in similar subdirectory tree => return path
                return out_path

        # Nothing found => raises IOError
        raise IOError("Cannot find file '{fname!s}' in '{dpath!s}' directory.".format(dpath=path, fname=fname))

    @staticmethod
    def last_modification_timestamp(file_path):
        """Returns the  POSIX timestamp corresponding to the last modification time of a given file"""
        fpath = FileUtil.locate_file(file_path)
        return int(os.path.getmtime(fpath))

    @staticmethod
    def _check_filename_is_string(filepath):
        if not Stringifiable.is_type_string(filepath):
            raise AttributeError("'filepath' attribute must be a valid string. "
                                 "Got '{fname!s}'.".format(fname=filepath))

    @staticmethod
    def _check_valid_file_type(ftype):
        if not isinstance(ftype, FileType):
            raise AttributeError("'file_type' must be a valid FileType enum value.")


class HDF5IOFile(object):
    def save_HDF5(self, *args, **kwargs):
        raise NotImplementedError()

    @classmethod
    def load_HDF5(cls, hdf5_file_path):
        raise NotImplementedError()

    def close(self):
        pass


__all__ = ["FileType", "FileUtil", "HDF5IOFile"]



# -*- coding: utf-8 -*-
# This file is part of the 'astrophysix' Python package.
#
# Copyright © Commissariat a l'Energie Atomique et aux Energies Alternatives (CEA)
#
#  FREE SOFTWARE LICENCING
#  -----------------------
# This software is governed by the CeCILL license under French law and abiding by the rules of distribution of free
# software. You can use, modify and/or redistribute the software under the terms of the CeCILL license as circulated by
# CEA, CNRS and INRIA at the following URL: "http://www.cecill.info". As a counterpart to the access to the source code
# and rights to copy, modify and redistribute granted by the license, users are provided only with a limited warranty
# and the software's author, the holder of the economic rights, and the successive licensors have only limited
# liability. In this respect, the user's attention is drawn to the risks associated with loading, using, modifying
# and/or developing or reproducing the software by the user in light of its specific status of free software, that may
# mean that it is complicated to manipulate, and that also therefore means that it is reserved for developers and
# experienced professionals having in-depth computer knowledge. Users are therefore encouraged to load and test the
# software's suitability as regards their requirements in conditions enabling the security of their systems and/or data
# to be ensured and, more generally, to use and operate it in the same conditions as regards security. The fact that
# you are presently reading this means that you have had knowledge of the CeCILL license and that you accept its terms.
#
#
# COMMERCIAL SOFTWARE LICENCING
# -----------------------------
# You can obtain this software from CEA under other licencing terms for commercial purposes. For this you will need to
# negotiate a specific contract with a legal representative of CEA.
#
"""

.. autoclass:: astrophysix.units.unit.Unit
   :members: UNKNOWN_PHYSICAL_TYPE, name, coeff, description, latex, dimensions, physical_type, appropriate_unit,
             equivalent_unit_list, express, from_name, identical, info, is_base_unit, iterate_units, __eq__, create_unit
   :undoc-members:

"""
from __future__ import print_function, unicode_literals, division
from future.builtins import dict, list
from numpy import isscalar, array, abs, isclose, argmin, log10
import sys
import warnings


class UnitError(Exception):
    pass


class Unit(object):
    """
    Dimensional physical unit class

    Parameters
    ----------
    name : :obj:`string`
        Unit name
    base_unit : :class:`~astrophysix.units.unit.Unit` instance
        Composite unit from which this instance should be initialised
    coeff  : :obj:`float`
        dimensionless value of the unit instance.
    dims : 8-:obj:`tuple` of :obj:`int`
        dimension of the unit object expressed in the international unit system (``kg``, ``m``, ``s``, ``K``, ``A``, ``mol``, ``rad``, ``cd``)
    descr : :obj:`string` or **None**
        Unit description
    latex: :obj:`string` or **None**
        Unit displayed name (latex format)

    Examples
    --------

        >>> cs_m_s = Unit(name="cs", coeff=340.0, dims=(0, 1, -1, 0, 0, 0, 0, 0), descr="sound speed unit")
        >>> print("sound speed = {v:g} m/h".format(v=cs_m_s.express(km/hour)))
        sound speed = 1224 km/h
        >>>
        >>> dens = Unit(name="Msun/kpc^3", base_unit=Msun/kpc**3, descr="Solar mass per cubic kiloparsec",
                        latex="{u1:s}.{u2:s}^{{-3}}".format(u1=Msun.latex, u2=kpc.latex))
        >>> print(dens)
        (6.76957356533e-29 m^-3.kg)
    """
    _phystype_unit_mapping = dict()
    UNKNOWN_PHYSICAL_TYPE = 'unknown'
    _unit_registry = dict()
    _unit_system = list(["kg", "m", "s", "K", "A", "mol", "rad", "cd"])

    def __new__(cls, name="", base_unit=None, coeff=1.0, dims=None, descr=None, latex=None):
        # Tries to create a named Unit instance with a name already present in the registry
        if name != "" and name in cls._unit_registry:
            # Get registry corresponding Unit
            u = cls._unit_registry[name]

            # Check if the new Unit is identical to the existing one in the registry
            duplicate_entry = False
            if base_unit is not None:
                if not isinstance(base_unit, Unit):
                    raise AttributeError("'base_unit' parameter must be a a 'Unit' instance.")

                # Different Unit with identical name ?
                if base_unit.coeff != u.coeff or ((base_unit.dimensions - u.dimensions) != 0).any():
                    duplicate_entry = True
            elif dims is not None:
                if len(dims) != len(cls._unit_system):
                    raise AttributeError("Wrong dimensions : must be a tuple of length " + str(len(cls._unit_system)))
                adims = array(dims, 'i')

                # Different Unit with identical name ?
                if ((u.dimensions - adims) != 0).any() or coeff != u.coeff:
                    duplicate_entry = True

            if duplicate_entry:
                warnings.warn("Warning ! a different unit with identical name '{uname:s}' already exist in Unit "
                              "registry. Registry Unit instance is used...".format(uname=name))
            else:  # Registry Unit is identical, check description/latex differences
                if descr is not None and descr != u.description:
                    warnings.warn("Warning ! a identical unit with name '{uname:s}' already exist in Unit registry with"
                                  " a different description. Registry Unit instance is used...".format(uname=name))
                elif latex is not None and latex != u.latex:
                    warnings.warn("Warning ! a identical unit with name '{uname:s}' already exist in Unit registry with"
                                  " a different LaTex formula. Registry Unit instance is used...".format(uname=name))
            return u

        return super(Unit, cls).__new__(cls)

    def __init__(self, name="", base_unit=None, coeff=1.0, dims=None, descr=None, latex=None):
        # Already initialised Unit instance (in the registry)
        if name != "" and name in self._unit_registry and self is self._unit_registry[name]:
            return

        super(Unit, self).__init__()

        # Unit name + laTex displayed name
        self._name = name
        self._latex = latex

        if base_unit is not None:
            if not isinstance(base_unit, Unit):
                raise AttributeError("'base_unit' parameter must be a a 'Unit' instance.")
            self._dimensions = array(base_unit._dimensions, 'i')
            self._coeff = base_unit._coeff
        elif dims is None:
            raise AttributeError("'dims' attribute is mandatory to define a 'Unit' object.")
        else:
            if len(dims) != len(Unit._unit_system):
                raise AttributeError("Wrong dimensions : must be a tuple of length " + str(len(Unit._unit_system)))
            self._dimensions = array(dims, 'i')
            self._coeff = coeff

        if descr is None:
            self._descr = ""
        else:
            self._descr = descr

    @classmethod
    def from_name(cls, unit_name):
        """
        Get a :class:`~astrophysix.units.unit.Unit` from its name in the ``astrophysix`` unit registry.

        Parameters
        ----------
        unit_name: :obj:`string`
            name of the unit to search.

        Raises
        ------
        AttributeError
            if `unit_name` attribute does not correspond to any unit in the ``astrophysix`` unit registry.
        """
        if unit_name not in cls._unit_registry:
            raise AttributeError("Unknown unit name '%s'" % unit_name)

        return cls._unit_registry[unit_name]

    @classmethod
    def create_unit(cls, name="", base_unit=None, coeff=1.0, dims=None, descr=None, latex=None):
        """
        Add a new Unit instance to the registry

        Parameters
        ----------
        name : :obj:`string`
            Unit name
        base_unit : :class:`~astrophysix.units.unit.Unit` instance
            Composite unit from which this instance should be initialised
        coeff  : :obj:`float`
            dimensionless value of the unit instance.
        dims : 8-:obj:`tuple` of :obj:`int`
            dimension of the unit object expressed in the international unit system (``kg``, ``m``, ``s``, ``K``, ``A``, ``mol``, ``rad``, ``cd``)
        descr : :obj:`string` or **None**
            Unit description
        latex: :obj:`string` or **None**
            Unit displayed name (latex format)

        Raises
        ------
        ValueError
            If the provided `name` already corresponds to a unit in the registry.
        """
        if name in cls._unit_registry:
            raise ValueError("{unm:s} Unit is already defined as {regunit!s}".format(unm=name,
                                                                                     regunit=cls._unit_registry[name]))

        u = Unit(name=name, base_unit=base_unit, coeff=coeff, dims=dims, descr=descr, latex=latex)
        cls._unit_registry[name] = u
        return u

    def __eq__(self, other):
        """
        Checks Unit instance equality

        Parameters
        ----------
        other: :class:`~astrophysix.units.unit.Unit`
            other unit instance to compare to

        Returns
        -------
        e: :obj:`bool`
            True if :attr:`Unit.coeff` and :attr:`Unit.dimensions` are identical, otherwise False.
        """
        if not isclose(self._coeff, other.coeff, rtol=1.0e-6):
            return False
        if ((self._dimensions - other.dimensions) != 0).any():
            return False
        return True

    def __ne__(self, other):
        """
        Not an implied relationship between "rich comparison" equality methods in Python 2.X but only in Python 3.X
        see https://docs.python.org/2.7/reference/datamodel.html#object.__ne__

        other: other instance to compare to
        """
        # --------------------------------------------- Python 2 ----------------------------------------------------- #
        #  There are no implied relationships among the comparison operators. The truth of x==y does not imply that
        #  x!=y is false. Accordingly, when defining __eq__(), one should also define __ne__() so that the operators
        #  will behave as expected.
        # --------------------------------------------- Python 3 ----------------------------------------------------- #
        # By default, __ne__() delegates to __eq__() and inverts the result unless it is NotImplemented. There are no
        # other implied relationships among the comparison operators, for example, the truth of (x<y or x==y) does not
        # imply x<=y.
        # ------------------------------------------------------------------------------------------------------------ #
        return not self.__eq__(other)

    def identical(self, other_unit):
        """
        Strict unit instance comparison method

        Parameters
        ----------
        other_unit: :class:`~astrophysix.units.unit.Unit`
            other unit to compare to.

        Returns
        -------
        e: :obj:`bool`
            True only if `other_unit` is equals to `self` AND has identical name/description/LaTex formula.
            Otherwise returns False.
        """
        # Strict value equality
        if self._coeff != other_unit.coeff or ((self._dimensions - other_unit.dimensions) != 0).any():
            return False

        if self._name != other_unit.name or self.latex != other_unit.latex or self._descr != other_unit.description:
            return False

        return True

    @property
    def name(self):
        """
        Unit name
        """
        return self._name

    @property
    def latex(self):
        """
        Unit displayed name (LaTex format)
        """
        if self._latex is not None:
            return self._latex
        return self._name

    @property
    def dimensions(self):
        """
        Unit dimension array
        """
        return self._dimensions

    @property
    def description(self):
        """
        Unit description
        """
        return self._descr

    @property
    def coeff(self):
        """
        Constant value of this unit
        """
        return self._coeff

    def is_base_unit(self):
        """
        Checks whether the Unit is a base SI Unit (kg, m, s, K, A, mol, rad, cd).

        Returns
        -------
        b: :obj:`bool`
            True only if unit is a base SI unit(kg, m, s, K, A, mol, rad, cd). Otherwise returns False.
        """
        if self._coeff != 1.0:
            return False

        msk = (self._dimensions != 0)
        n = msk.sum()
        if n == 0:  # Dimensionless unit
            return True
        elif n == 1:
            if self._dimensions[msk][0] == 1:
                return True

        return False

    def info(self):
        """
        Print information about this unit. If any, print the name and description of this unit, then print the value of
        this unit and the list of equivalent unit contained in the built-in unit registry associated with their
        conversion factor.

        Example
        -------
            >>> U.kpc.info()
            Unit : kpc
            ----------
            Kiloparsec
            Value
            -----
            3.0856775814671917e+19 m
            Equivalent units
            ----------------
            * m            :    3.24078e-20 kpc
            * um           :    3.24078e-26 kpc
            * mm           :    3.24078e-23 kpc
            * cm           :    3.24078e-22 kpc
            * nm           :    3.24078e-29 kpc
            * km           :    3.24078e-17 kpc
            * Angstrom     :    3.24078e-30 kpc
            * au           :    4.84814e-09 kpc
            * pc           :          0.001 kpc
            * Mpc          :           1000 kpc
            * Gpc          :          1e+06 kpc
            * Rsun         :    2.25399e-11 kpc
            * ly           :    0.000306601 kpc
        """
        s = ""
        if self._name != "":
            title = "Unit : {uname:s}".format(uname=self._name)
            l = len(title)
            s += title + "\n" + (l * "-") + "\n"

        # Description
        if len(self._descr) > 0:
            s += self._descr + "\n\n"

        s += "Value\n" \
             "-----\n" + str(self._coeff) + " " + self._decompose_base_units() + "\n\n"

        s += "Equivalent units\n" \
             "----------------\n" + ""

        for u in self.equivalent_unit_list():
            s += " * {equname:12s} : {f:14g} {uname:s}\n".format(equname=u.name, f=u.express(self), uname=self._name)

        print(s)

    def equivalent_unit_list(self):
        """
        Get the equivalent unit list (with same physical type)

        Example
        -------
            >>> print(U.kg.equivalent_unit_list())
            [g : (0.001 kg), t : (1000 kg), mH : (1.66e-27 kg), Msun : (1.9889e+30 kg), Mearth : (5.9722e+24 kg)]
        """
        l = list()
        t = self.physical_type
        for u in self._unit_registry.values():
            if self == u:
                continue
            if u.physical_type == t:
                l.append(u)

        return l

    def appropriate_unit(self, nearest_log10=1.0):
        """
        Try to find the better suited unit (among available equivalent units to represent this unit).

        Parameters
        ----------
        nearest_log10: :obj:`float`
            log of the nearest value to round to. Default 1.0.

        Example
        -------
            >>> u = 2426.2 * U.ly
            >>> bv, bu = u.appropriate_unit()
            >>> print("Appropriate unit : 2426.2 ly = {v:g} {bu:s}".format(v=bv, bu=bu.name))
            Appropriate unit : 2426.2 ly = 0.743876 kpc
        """
        eq_units = self.equivalent_unit_list()
        if len(eq_units) == 0:
            return self

        tlist = [(eq_unit, self.express(eq_unit)) for eq_unit in eq_units]
        imin = argmin(array([abs(log10(t[1]) - nearest_log10) for t in tlist]))
        best_unit, ph_val = tlist[imin]
        return ph_val, best_unit

    @classmethod
    def create_phystype(cls, unit, phys_type_name):
        """
        Adds a new physical unit mapping.

        Parameters
        ----------
        unit : :class:`~astrophysix.units.unit.Unit`
            The unit to map from.

        phys_type_name : :obj:`string`
            The physical quantity name of the unit.
        """
        d = tuple(unit.dimensions)
        if d in cls._phystype_unit_mapping:
            raise ValueError("{u!s} is already defined as '{ptype_name:s}' physical quantity "
                             "type".format(u=cls._phystype_unit_mapping[d], ptype_name=phys_type_name))
        cls._phystype_unit_mapping[d] = phys_type_name

    @property
    def physical_type(self):
        """
        Get the unit physical type (dimensioned physical quantity).

        Returns
        -------
        t : :obj:`string`
            The name of the physical quantity, or :attr:`Unit.UNKNOWN_PHYSICAL_TYPE` if the physical quantity is unknown.
        """
        d = tuple(self._dimensions)
        return self._phystype_unit_mapping.get(d, self.UNKNOWN_PHYSICAL_TYPE)

    def __div__(self, other):
        r"""
        Returns the result of the division of `self` by `other`

        """
        if isscalar(other):
            newdims = self._dimensions
            newval = self._coeff / other
            return Unit(coeff=newval, dims=newdims)
        elif isinstance(other, Unit):
            newdims = self._dimensions - other._dimensions
            newval = self._coeff / other._coeff
            return Unit(coeff=newval, dims=newdims)
        else:
            raise UnitError("Unable to divide a Unit instance by something which is neither a Unit object nor a "
                            "scalar.")

    def __truediv__(self, other):
        r"""
        Returns the result of the future division of `self` by `other`

        """
        return self.__div__(other)

    def __rdiv__(self, other):
        r"""
        Returns the result of the division of `other` by `self`

        """
        if isscalar(other):
            newdims = -self._dimensions
            newval = other / self._coeff
            return Unit(coeff=newval, dims=newdims)
        elif isinstance(other, Unit):
            newdims = other._dimensions - self._dimensions
            newval = other._coeff / self._coeff
            return Unit(coeff=newval, dims=newdims)
        else:
            raise UnitError("Unable to divide something which is neither a Unit object nor a scalar by a Unit "
                            "instance.")

    def __rtruediv__(self, other):
        r"""
        Returns the result of the future division of `other` by `self`

        """
        return self.__rdiv__(other)

    def __mul__(self, other):
        r"""
        Returns the result of the multiplication of `self` with `other`

        """
        if isscalar(other):
            newdims = self._dimensions
            newval = self._coeff * other
            return Unit(coeff=newval, dims=newdims)
        elif isinstance(other, Unit):
            newdims = self._dimensions + other._dimensions
            newval = self._coeff * other._coeff
            return Unit(coeff=newval, dims=newdims)
        else:
            raise UnitError("Unable to multiply a Unit instance by something which is neither a Unit object nor a "
                            "scalar.")

    def __rmul__(self, other):
        r"""
        Returns the result of the multiplication of `other` with `self`

        """
        return self * other

    def __pow__(self, n):
        r"""
        Returns the result of the exponentiation of `self` by `n`

        """
        return Unit(coeff=self._coeff ** n, dims=n * self._dimensions)

    def _decompose_base_units(self):
        s = ""

        d = array(self._dimensions)
        for i in range(len(Unit._unit_system)):
            if d[i] != 0:
                if (d[0:i] != 0).any():
                    s += "."
                s += Unit._unit_system[i]
                if d[i] != 1:
                    s += "^" + str(d[i])
        return s

    if sys.version_info[0] >= 3:  # Python 3
        def __str__(self):
            return self.__unicode__()
    else:  # Python 2
        def __str__(self):
            return self.__unicode__().encode('utf8')

    def __unicode__(self):
        strrep = ""

        # Name & value
        if len(self._name) > 0:
            strrep += self._name + " : "

        strrep += "({coeff:g} {base_units:s})".format(coeff=self._coeff, base_units=self._decompose_base_units())
        return strrep

    def __repr__(self):
        return self.__str__()

    @classmethod
    def iterate_units(cls, phys_type=None):
        """
        Unit iterator method. Iterates over all units in the ``astrophysix`` unit registry.

        Parameters
        ----------
        phys_type: :obj:`string`
            Name of the physical quantity type of the units to iterate over. Default **None** (all physical quantities).

        Yields
        ------
        u: :class:`~astrophysix.units.unit.Unit`
            unit of the required physical quantity type, if any given.
        """
        if phys_type is None:
            for u in cls._unit_registry.values():
                yield u
        else:
            if phys_type not in cls._phystype_unit_mapping.values():
                raise AttributeError("Unknown physical quantity type '{pt!s}'.".format(pt=phys_type))
            for u in cls._unit_registry.values():
                if u.physical_type == phys_type:
                    yield u

    def express(self, unit):
        """
        Unit conversion method. Gives the conversion factor of this :class:`~astrophysix.units.unit.Unit`
        expressed into another (dimension-compatible) given :class:`~astrophysix.units.unit.Unit`.

        Checks that :

        * the **unit** param. is also a :class:`~astrophysix.units.unit.Unit` instance
        * the **unit** param. is dimension-compatible.

        Parameters
        ----------
        unit : :class:`~astrophysix.units.unit.Unit`
            unit in which the conversion is made

        Returns
        -------
        fact : :obj:`float`
            conversion factor of this unit expressed in **unit**

        Examples
        --------
        * Conversion of a kpc expressed in light-years :

            >>> factor = kpc.express(ly)
            >>> print("1 kpc = {fact:f} ly".format(fact=factor))
            1 kpc = 3261.563777 ly

        * Conversion of :math:`1 M_{\odot}` into kpc/Myr :

            >>> print(Msun.express(kpc/Myr))
            UnitError: Incompatible dimensions between :
            - Msun : (1.9889e+30 kg) (type: mass) and
            - (977792 m.s^-1) (type: velocity)
        """
        if not isinstance(unit, Unit):
            raise AttributeError("'unit' attribute must be a 'Unit' instance.")

        dmax = len(Unit._unit_system) - 1
        if tuple(self._dimensions[0:dmax]) != tuple(unit._dimensions[0:dmax]):
            raise UnitError("Incompatible dimensions between :\n"
                            "- {me!s} (type: {ptype:s}) and\n"
                            "- {unit!s} (type: {uptype:s})".format(me=self, ptype=self.physical_type, unit=unit,
                                                                   uptype=unit.physical_type))

        return (self / unit)._coeff


__all__ = ["Unit", "UnitError"]



# -*- coding: utf-8 -*-
# This file is part of the 'astrophysix' Python package.
#
# Copyright © Commissariat a l'Energie Atomique et aux Energies Alternatives (CEA)
#
#  FREE SOFTWARE LICENCING
#  -----------------------
# This software is governed by the CeCILL license under French law and abiding by the rules of distribution of free
# software. You can use, modify and/or redistribute the software under the terms of the CeCILL license as circulated by
# CEA, CNRS and INRIA at the following URL: "http://www.cecill.info". As a counterpart to the access to the source code
# and rights to copy, modify and redistribute granted by the license, users are provided only with a limited warranty
# and the software's author, the holder of the economic rights, and the successive licensors have only limited
# liability. In this respect, the user's attention is drawn to the risks associated with loading, using, modifying
# and/or developing or reproducing the software by the user in light of its specific status of free software, that may
# mean that it is complicated to manipulate, and that also therefore means that it is reserved for developers and
# experienced professionals having in-depth computer knowledge. Users are therefore encouraged to load and test the
# software's suitability as regards their requirements in conditions enabling the security of their systems and/or data
# to be ensured and, more generally, to use and operate it in the same conditions as regards security. The fact that
# you are presently reading this means that you have had knowledge of the CeCILL license and that you accept its terms.
#
#
# COMMERCIAL SOFTWARE LICENCING
# -----------------------------
# You can obtain this software from CEA under other licencing terms for commercial purposes. For this you will need to
# negotiate a specific contract with a legal representative of CEA.
#
"""

.. autoclass:: astrophysix.simdm.protocol.InputParameter
    :members:
    :undoc-members:
    :inherited-members:
    :exclude-members: open_h5file, is_type_string, hsp_save_to_h5, cast_string, hsp_load_from_h5,
                      INVALID_ALIAS_ERROR_MESSAGE, VALID_ALIAS_REGEX

"""
from __future__ import absolute_import, unicode_literals
from future.builtins import str, list, dict
import logging

from astrophysix.simdm.utils import GalacticaValidityCheckMixin
from astrophysix.utils.persistency import Hdf5StudyPersistent
from astrophysix.utils.strings import Stringifiable

log = logging.getLogger("astrophysix.simdm")


class InputParameter(Hdf5StudyPersistent, GalacticaValidityCheckMixin, Stringifiable):
    """
    Protocol input parameter

    Parameters
    ----------
    name: :obj:`string`
        input parameter name (mandatory)
    key: :obj:`string`
        input parameter configuration key
    description: :obj:`string`
        input parameter description
    """
    def __init__(self, **kwargs):
        super(InputParameter, self).__init__(**kwargs)

        self._key = ""
        self._name = ""
        self._description = ""

        if "name" not in kwargs:
            raise AttributeError("Input parameter 'name' attribute is not defined (mandatory).")
        self.name = kwargs["name"]

        if "key" in kwargs:
            self.key = kwargs["key"]

        if "description" in kwargs:
            self.description = kwargs["description"]

    def __eq__(self, other):
        """
        InputParameter comparison method

        other: :class:`~astrophysix.simdm.protocol.InputParameter`
            input parameter to compare to
        """
        if not super(InputParameter, self).__eq__(other):
            return False

        if self._name != other.name:
            return False

        if self._key != other.key:
            return False

        if self._description != other.description:
            return False

        return True

    @property
    def key(self):
        """Input parameter configuration key"""
        return self._key

    @key.setter
    def key(self, new_key):
        try:
            self._key = Stringifiable.cast_string(new_key)
        except TypeError:
            raise AttributeError("Input parameter 'key' property is not a valid string.")

    @property
    def name(self):
        """Input parameter name"""
        return self._name

    @name.setter
    def name(self, new_name):
        try:
            self._name = Stringifiable.cast_string(new_name, valid_empty=False)
        except TypeError:
            raise AttributeError("Input parameter 'name' property is not a valid (non empty) string.")

    @property
    def description(self):
        """Input parameter description"""
        return self._description

    @description.setter
    def description(self, new_descr):
        try:
            self._description = Stringifiable.cast_string(new_descr)
        except TypeError:
            raise AttributeError("Input parameter 'description' property is not a valid string.")

    @classmethod
    def _hsp_valid_attributes(cls):
        """List of valid kwargs in __init__() method"""
        return ["name", "key", "description"]

    def _hsp_write(self, h5group, **kwargs):
        """
        Serialize an InputParameter object into a HDF5 file.

        Parameters
        ----------
        h5group: ``h5py.Group``
            Main group to write the InputParameter into.
        kwargs: ``dict``
            keyword argument dictionary.
        """
        # Call to parent class _hsp_write() : write UUID, etc.
        super(InputParameter, self)._hsp_write(h5group, **kwargs)

        # Write input parameter key
        self._hsp_write_attribute(h5group, ('key', self._key), **kwargs)

        # Write input parameter name, if defined
        self._hsp_write_attribute(h5group, ('name', self._name), **kwargs)

        # Write input parameter description, if defined
        self._hsp_write_attribute(h5group, ('description', self._description), **kwargs)

    @classmethod
    def _hsp_read(cls, h5group, version, dependency_objdict=None):
        """
        Read an InputParameter object from a HDF5 file (*.h5).

        Parameters
        ----------
        h5group: ``h5py.Group``
            Main group to read the object from.
        version: ``int``
            version of the object to read.
        dependency_objdict: ``dict``
            dependency object dictionary. Default None

        Returns
        -------
        exp: ``InputParameter``
            Read InputParameter instance
        """
        # Handle different versions here

        # Fetch Hdf5StudyPersistent object UUID
        uid = super(InputParameter, cls)._hsp_read(h5group, version, dependency_objdict=dependency_objdict)

        # Read input parameter key
        ip_name = cls._hsp_read_attribute(h5group, 'name', "input parameter name")

        # Create input parameter object
        ip = cls(uid=uid, name=ip_name)

        # Read input parameter key, if defined
        ip_key = cls._hsp_read_attribute(h5group, 'key', "input parameter key", raise_error_if_not_found=False)
        if ip_key is not None:
            ip.key = ip_key

        # Read input parameter description, if defined
        ip_descr = cls._hsp_read_attribute(h5group, 'description', "input parameter description",
                                           raise_error_if_not_found=False)
        if ip_descr is not None:
            ip.description = ip_descr

        return ip

    def galactica_validity_check(self, **kwargs):
        """
        Perform validity checks on this instance and eventually log warning messages.

        Parameters
        ----------
        kwargs: :obj:`dict`
            keyword arguments (optional)
        """
        # Check input parameter name
        if len(self._name) == 0:
            log.warning("{p!s} Galactica input parameter name is missing.".format(p=self))
        elif len(self._name) > 64:
            log.warning("{p!s} Galactica input parameter name is too long (max. 64 characters).".format(p=self))

        # Check input parameter key
        if len(self._key) > 16:
            log.warning("{p!s} Galactica input parameter key is too long (max. 16 characters).".format(p=self))

    def __unicode__(self):
        """
        String representation of the instance
        """
        s = "[{ipk:s}]".format(ipk=self._key)
        if len(self._name) > 0:
            s += " '{ipname:s}'".format(ipname=self._name)
        s += " input parameter"
        return s


__all__ = ["InputParameter"]



# -*- coding: utf-8 -*-
# This file is part of the 'astrophysix' Python package.
#
# Copyright © Commissariat a l'Energie Atomique et aux Energies Alternatives (CEA)
#
#  FREE SOFTWARE LICENCING
#  -----------------------
# This software is governed by the CeCILL license under French law and abiding by the rules of distribution of free
# software. You can use, modify and/or redistribute the software under the terms of the CeCILL license as circulated by
# CEA, CNRS and INRIA at the following URL: "http://www.cecill.info". As a counterpart to the access to the source code
# and rights to copy, modify and redistribute granted by the license, users are provided only with a limited warranty
# and the software's author, the holder of the economic rights, and the successive licensors have only limited
# liability. In this respect, the user's attention is drawn to the risks associated with loading, using, modifying
# and/or developing or reproducing the software by the user in light of its specific status of free software, that may
# mean that it is complicated to manipulate, and that also therefore means that it is reserved for developers and
# experienced professionals having in-depth computer knowledge. Users are therefore encouraged to load and test the
# software's suitability as regards their requirements in conditions enabling the security of their systems and/or data
# to be ensured and, more generally, to use and operate it in the same conditions as regards security. The fact that
# you are presently reading this means that you have had knowledge of the CeCILL license and that you accept its terms.
#
#
# COMMERCIAL SOFTWARE LICENCING
# -----------------------------
# You can obtain this software from CEA under other licencing terms for commercial purposes. For this you will need to
# negotiate a specific contract with a legal representative of CEA.
#
"""

.. autoclass:: astrophysix.simdm.datafiles.image.PngImageFile
    :members: filename, last_modified, load_file, save_to_disk, pil_image, raw_file_data
    :show-inheritance: AssociatedFile

.. autoclass:: astrophysix.simdm.datafiles.image.JpegImageFile
    :members: filename, last_modified, load_file, save_to_disk, pil_image, raw_file_data
    :show-inheritance: AssociatedFile

"""
from __future__ import print_function, unicode_literals, division, absolute_import  # Python 2 and 3 compatibility
from future.builtins import str, list, int
import logging
import io
from PIL import Image

from astrophysix.utils.file import FileType
from .file import AssociatedFile


log = logging.getLogger("astrophysix.simdm")


class ImageFile(AssociatedFile):
    @property
    def pil_image(self):
        """
        Pillow image (JPEG/PNG) image property getter. Implements lazy I/O.
        """
        # Force read from HDF5 file if necessary
        self._hsp_lazy_read()

        return Image.open(io.BytesIO(self._raw_file_data.tobytes()))

    def __unicode__(self):
        """
        String representation of the ImageFile instance
        """
        s = "[{fname:s}] datafile associated {ft:s} image file".format(fname=self._filename, ft=self.FILE_TYPE.alias)
        return s


class PngImageFile(ImageFile):
    """
    Datafile associated :attr:`~astrophysix.utils.file.FileType.PNG_FILE` image file class.
    """
    FILE_TYPE = FileType.PNG_FILE


class JpegImageFile(ImageFile):
    """
    Datafile associated :attr:`~astrophysix.utils.file.FileType.JPEG_FILE` image file class.
    """
    FILE_TYPE = FileType.JPEG_FILE


__all__ = ["ImageFile", "JpegImageFile", "PngImageFile"]



# -*- coding: utf-8 -*-
# This file is part of the 'astrophysix' Python package.
#
# Copyright © Commissariat a l'Energie Atomique et aux Energies Alternatives (CEA)
#
#  FREE SOFTWARE LICENCING
#  -----------------------
# This software is governed by the CeCILL license under French law and abiding by the rules of distribution of free
# software. You can use, modify and/or redistribute the software under the terms of the CeCILL license as circulated by
# CEA, CNRS and INRIA at the following URL: "http://www.cecill.info". As a counterpart to the access to the source code
# and rights to copy, modify and redistribute granted by the license, users are provided only with a limited warranty
# and the software's author, the holder of the economic rights, and the successive licensors have only limited
# liability. In this respect, the user's attention is drawn to the risks associated with loading, using, modifying
# and/or developing or reproducing the software by the user in light of its specific status of free software, that may
# mean that it is complicated to manipulate, and that also therefore means that it is reserved for developers and
# experienced professionals having in-depth computer knowledge. Users are therefore encouraged to load and test the
# software's suitability as regards their requirements in conditions enabling the security of their systems and/or data
# to be ensured and, more generally, to use and operate it in the same conditions as regards security. The fact that
# you are presently reading this means that you have had knowledge of the CeCILL license and that you accept its terms.
#
#
# COMMERCIAL SOFTWARE LICENCING
# -----------------------------
# You can obtain this software from CEA under other licencing terms for commercial purposes. For this you will need to
# negotiate a specific contract with a legal representative of CEA.
#
from __future__ import print_function, unicode_literals, division, absolute_import  # Python 2 and 3 compatibility
from future.builtins import str
import os
import sys
import shutil
import uuid
import logging

if sys.version_info.major == 2:
    import pytz

log = logging.getLogger("astrophysix.simdm")

from .project import Project
from . import ProjectCategory
from ..utils.persistency import Hdf5StudyPersistent, Hdf5StudyFileMode
from ..utils.dt_utils import DatetimeUtil
from ..utils.file import FileUtil, HDF5IOFile


class SimulationStudy(HDF5IOFile):
    """
    HDF5 simulation study file for Project tree structure persistency

    Parameters
    ----------
    project: :class:`~astrophysix.simdm.Project`
        study main project
    """
    def __init__(self, project=None):
        super(SimulationStudy, self).__init__()

        # Datetimes
        self._created = DatetimeUtil.utc_now()
        self._last_modified = DatetimeUtil.utc_now()

        # Study UUID
        self._uid = uuid.uuid4()

        # Study file paths
        self._study_filepath = None
        self._temp_study_filepath = FileUtil.new_temp_filepath("study")

        # Study project
        if project is None:
            # Create empty dummy project
            self._project = Project(category=ProjectCategory.Cosmology, project_title="My new project")
        else:
            # Check that project is a valid Project instance
            if not isinstance(project, Project):
                err_msg = "{cname:s} 'project' attribute is not a valid Project " \
                          "object.".format(cname=self.__class__.__name__)
                log.error(err_msg)
                raise AttributeError(err_msg)
            self._project = project


    @property
    def study_filepath(self):
        """Simulation study HDF5 file path"""
        return self._study_filepath

    @property
    def project(self):
        """Study main project"""
        return self._project

    @property
    def uid(self):
        """Study UUID"""
        return self._uid

    @property
    def creation_time(self):
        """Study creation date/time (:class:`datetime.datetime`)."""
        return self._created

    @property
    def last_modification_time(self):
        """Study last modification date/time (:class:`datetime.datetime`)."""
        return self._last_modified

    def save_HDF5(self, study_fname=None, dry_run=False, callback=None, galactica_checks=False):
        """
        Save the SimulationStudy into a HDF5 (\\*.h5) file

        Parameters
        ----------
        study_fname: :obj:`string`
            Simulation study HDF5 filename.
        dry_run: :obj:`bool`
            perform a dry run ? Default False.
        callback: :obj:`callable`
            method to execute upon saving each item of the study.
        galactica_checks: :obj:`bool`
            Perform Galactica database validity checks and display warning in case of invalid content for upload on
            Galactica. Default False (quiet mode).
        """
        if study_fname is None and self._study_filepath is None:
            # No file path provided : should never happen
            err_msg = "No filename provided. Please provide a HDF5 filename to save the study."
            log.error(err_msg)
            raise AttributeError(err_msg)

        # If Galactica database validity checks is enabled, perform a full check of the project before saving it.
        if galactica_checks:
            self._project.galactica_validity_check()

        # Save project into the temp. file
        if self._study_filepath is None:  # New file, study was never saved
            h5f, h5group, close_when_done = Hdf5StudyPersistent.open_h5file(self._temp_study_filepath,
                                                                            mode=Hdf5StudyFileMode.NEW_WRITE)
            try:
                Hdf5StudyPersistent._hsp_write_object(h5group, "PROJECT", self._project, new_file=True, dry_run=dry_run,
                                                      callback_func=callback, from_project=True)

                ctms = DatetimeUtil.utc_to_timestamp(self._created)
                h5group.attrs["ObjectClass"] = "Hdf5_Persistent_Simulation_Study"
                h5group.attrs["creation_time"] = ctms
                self._last_modified = DatetimeUtil.utc_now()
                mtms = DatetimeUtil.utc_to_timestamp(self._last_modified)
                h5group.attrs["last_modif_time"] = mtms
                h5group.attrs["study_version"] = 1
                h5group.attrs["study_uid"] = "{uid!s}".format(uid=self._uid)
            except Exception:
                raise
            finally:
                if close_when_done and h5f is not None:
                    h5f.close()

        else:  # Old study (already saved)
            alt_temp_file = FileUtil.new_temp_filepath("study")
            shutil.copy(self._temp_study_filepath, alt_temp_file)
            h5f, h5group, close_when_done = Hdf5StudyPersistent.open_h5file(alt_temp_file,
                                                                            mode=Hdf5StudyFileMode.APPEND)
            try:
                Hdf5StudyPersistent._hsp_write_object(h5group, "PROJECT", self._project, new_file=False,
                                                      dry_run=dry_run, callback_func=callback, from_project=True)

                mtms = DatetimeUtil.utc_to_timestamp(self._last_modified)
                h5group.attrs["last_modif_time"] = mtms
            except Exception:
                raise
            finally:
                if close_when_done and h5f is not None:
                    h5f.close()

            if not dry_run:
                # If no exception was raised, switch temp file
                if os.path.exists(self._temp_study_filepath):
                    os.remove(self._temp_study_filepath)
                self._temp_study_filepath = alt_temp_file
            else:
                # Delete temporary file (used for dry run)
                os.remove(alt_temp_file)

        if not dry_run:
            if study_fname is None:
                dest_path = self._study_filepath  # Keep old study file name
            else:
                dest_path = study_fname  # Change destination study filename

            # Move temporary file to the requested study HDF5 file path
            if os.path.exists(dest_path):  # Delete first the already existing file
                os.remove(dest_path)
            shutil.copy(self._temp_study_filepath, dest_path)

            # Set current study file path
            self._study_filepath = dest_path

    @classmethod
    def load_HDF5(cls, study_file_path):
        """
        Loads a new or existing SimulationStudy from a HDF5 (\\*.h5) file

        Parameters
        ----------
        study_file_path: :obj:`string`
            SimulationStudy HDF5 (existing) file path

        Returns
        -------
        study: :class:`~astrophysix.simdm.SimulationStudy`
            Study loaded from HDF5 file.
        """
        if not os.path.isfile(study_file_path):
            raise AttributeError("Cannot find file '{fname:s}'".format(fname=study_file_path))

        temp_loaded_file = FileUtil.new_temp_filepath("study")
        shutil.copy(study_file_path, temp_loaded_file)
        h5f, h5group, close_when_done = Hdf5StudyPersistent.open_h5file(temp_loaded_file,
                                                                        mode=Hdf5StudyFileMode.READ_ONLY)
        try:
            # Check that we are actually reading a Simulation study object from a HDF5 file
            if "ObjectClass" not in h5group.attrs or h5group.attrs["ObjectClass"] != "Hdf5_Persistent_Simulation_Study":
                err_msg = "HDF5 file does not contain a simulation study !"
                log.error(err_msg)
                raise IOError(err_msg)

            # Read study UUID
            if "study_uid" not in h5group.attrs:
                err_msg = "Cannot find study UUID attribute in '{path!s}'.".format(path=h5group.name)
                log.error(err_msg)
                raise IOError(err_msg)

            # Read study creation/last modification times
            if "creation_time" not in h5group.attrs or "last_modif_time" not in h5group.attrs:
                err_msg = "Cannot find study creation/modification time attributes in " \
                          "'{path!s}'.".format(path=h5group.name)
                log.error(err_msg)
                raise IOError(err_msg)

            if "PROJECT" not in h5group:
                err_msg = "Missing '/PROJECT' group in HDF5 simulation study file."
                log.error(err_msg)
                raise IOError(err_msg)
            proj_group = h5group["PROJECT"]
            p = Project.hsp_load_from_h5(proj_group)
            study = cls(project=p)
            study._temp_study_filepath = temp_loaded_file
            study._study_filepath = study_file_path
            study._uid = uuid.UUID(h5group.attrs["study_uid"])
            study._created = DatetimeUtil.utc_from_timestamp(h5group.attrs["creation_time"])
            study._last_modified = DatetimeUtil.utc_from_timestamp(h5group.attrs["last_modif_time"])
        except Exception:
            raise
        finally:
            if close_when_done and h5f is not None:
                h5f.close()

        return study

    def close(self):
        self._delete()

    def _delete(self):
        """
        Clean the study temporary HDF5 file (in /tmp).
        """
        if os.path.exists(self._temp_study_filepath):
            os.remove(self._temp_study_filepath)


__all__ = ["SimulationStudy"]



# -*- coding: utf-8 -*-
# This file is part of the 'astrophysix' Python package.
#
# Copyright © Commissariat a l'Energie Atomique et aux Energies Alternatives (CEA)
#
#  FREE SOFTWARE LICENCING
#  -----------------------
# This software is governed by the CeCILL license under French law and abiding by the rules of distribution of free
# software. You can use, modify and/or redistribute the software under the terms of the CeCILL license as circulated by
# CEA, CNRS and INRIA at the following URL: "http://www.cecill.info". As a counterpart to the access to the source code
# and rights to copy, modify and redistribute granted by the license, users are provided only with a limited warranty
# and the software's author, the holder of the economic rights, and the successive licensors have only limited
# liability. In this respect, the user's attention is drawn to the risks associated with loading, using, modifying
# and/or developing or reproducing the software by the user in light of its specific status of free software, that may
# mean that it is complicated to manipulate, and that also therefore means that it is reserved for developers and
# experienced professionals having in-depth computer knowledge. Users are therefore encouraged to load and test the
# software's suitability as regards their requirements in conditions enabling the security of their systems and/or data
# to be ensured and, more generally, to use and operate it in the same conditions as regards security. The fact that
# you are presently reading this means that you have had knowledge of the CeCILL license and that you accept its terms.
#
#
# COMMERCIAL SOFTWARE LICENCING
# -----------------------------
# You can obtain this software from CEA under other licencing terms for commercial purposes. For this you will need to
# negotiate a specific contract with a legal representative of CEA.
#
"""

.. autoclass:: astrophysix.simdm.services.process.DataProcessingService
    :members:
    :undoc-members:
    :inherited-members:
    :exclude-members: open_h5file, is_type_string, hsp_save_to_h5, cast_string, hsp_load_from_h5,
                      INVALID_ALIAS_ERROR_MESSAGE, VALID_ALIAS_REGEX


.. autoclass:: astrophysix.simdm.services.process.CatalogDataProcessingService
    :members:
    :undoc-members:
    :inherited-members:
    :exclude-members: open_h5file, is_type_string, hsp_save_to_h5, cast_string, hsp_load_from_h5,
                      INVALID_ALIAS_ERROR_MESSAGE, VALID_ALIAS_REGEX


.. autoclass:: astrophysix.simdm.services.process.CatalogFieldBinding
    :members:
    :undoc-members:
    :inherited-members:
    :exclude-members: open_h5file, is_type_string, hsp_save_to_h5, cast_string, hsp_load_from_h5,
                      INVALID_ALIAS_ERROR_MESSAGE, VALID_ALIAS_REGEX

"""

from .process import DataProcessingService, CatalogDataProcessingService, CatalogFieldBinding


__all__ = ["DataProcessingService", "CatalogDataProcessingService", "CatalogFieldBinding"]
